#  Copyright (c) Oba-group 
#  Distributed under the terms of the MIT License.

from pathlib import Path

from vise.chempotdiag.gas import (
    FundamentalFrequencies, ShomateThermodynamicsFunction, Gas)

from vise.config import ROOM_TEMPERATURE, REFERENCE_PRESSURE
from vise.util.testing import ViseTest

molecules_dir = Path(__file__).parent / ".." / "molecules"


class TestFundamentalFrequencies(ViseTest):

    def test_from_yaml(self):
        nh3_yaml = molecules_dir / "NH3" / "fundamental_frequencies.yaml"

        nh3 = FundamentalFrequencies.from_yaml(nh3_yaml)
        expected = [3506, 1022, 3577, 1691]
        actual = [f.wave_number for f in nh3.frequencies]
        self.assertListEqual(expected, actual)

        expected = [1, 1, 2, 2]
        actual = [f.degeneration for f in nh3.frequencies]
        self.assertListEqual(expected, actual)


class TestShomateThermodynamicsFunction(ViseTest):

    def setUp(self):
        o2_file = molecules_dir / "O2" / "shomate_nist.dat"
        no2_file = molecules_dir / "NO2" / "shomate_nist.dat"
        self.o2 = ShomateThermodynamicsFunction.from_nist_table(o2_file)
        self.no2 = ShomateThermodynamicsFunction.from_nist_table(no2_file)

    def test_heat_capacity(self):
        """
        Reference:
        Malcom W. Chase Jr. NIST-JANAF Thermodynamical Tables Fourth Edition,
        J. Phys. Chem. Ref. Data, Monograph 9
        https://srd.nist.gov/JPCRD/jpcrdM9.pdf
        """
        self.assertAlmostEqual(self.o2.heat_capacity(ROOM_TEMPERATURE),
                               29.376, 1)
        self.assertAlmostEqual(self.o2.heat_capacity(400), 30.106, 1)
        self.assertAlmostEqual(self.o2.heat_capacity(1500), 36.544, 1)
        self.assertAlmostEqual(self.o2.heat_capacity(5500), 43.426, 1)

    def test_standard_enthalpy(self):
        """
        Expected value:
        Malcom W. Chase Jr. NIST-JANAF Thermodynamical Tables Fourth Edition,
        J. Phys. Chem. Ref. Data, Monograph 9
        https://srd.nist.gov/JPCRD/jpcrdM9.pdf
        """
        self.assertAlmostEqual(self.o2.standard_enthalpy(ROOM_TEMPERATURE),
                               0, 1)
        self.assertAlmostEqual(self.o2.standard_enthalpy(400), 3.025 / 1000, 1)
        self.assertAlmostEqual(self.o2.standard_enthalpy(1500), 40.599 / 1000, 1)
        self.assertAlmostEqual(self.o2.standard_enthalpy(5500),
                               202.267 / 1000, 1)

    def test_standard_entropy(self):
        """
        Expected value:
        Malcom W. Chase Jr. NIST-JANAF Thermodynamical Tables Fourth Edition,
        J. Phys. Chem. Ref. Data, Monograph 9
        https://srd.nist.gov/JPCRD/jpcrdM9.pdf
        """
        self.assertAlmostEqual(self.o2.standard_entropy(ROOM_TEMPERATURE),
                               205.147, 1)
        self.assertAlmostEqual(self.o2.standard_entropy(400), 213.871, 1)
        self.assertAlmostEqual(self.o2.standard_entropy(1500), 258.068, 1)
        self.assertAlmostEqual(self.o2.standard_entropy(5500), 309.639, 1)

    def test_pressure_term(self):
        self.assertAlmostEqual(self.o2.r_ln_p_p0(REFERENCE_PRESSURE), 0)

    def test_zero_point(self):
        self.assertAlmostEqual(self.o2.enthalpy_zero, -8.683 / 1000)


class TestGas(ViseTest):

    def setUp(self):
        self.o2 = Gas.O2

    def test_read_properties(self):
        self.assertEqual(self.o2.properties["incar"]["NUPDOWN"], 2)

    def test_str(self):
        self.assertEqual(str(self.o2), "O2")

    def test_thermodynamics(self):
        """
        Just check Gas class's thermodynamical quantity is equal to
        ShomateThermodynamicsFunction.
        In TestShomateThermodynamicalFunction, these values are compared to
        those from NIST=JANAF Thermodynamical Tables.

        """
        o2_file = molecules_dir / "O2" / "shomate_nist.dat"
        o2_shomate = ShomateThermodynamicsFunction.from_nist_table(o2_file)

        self.assertEqual(self.o2.heat_capacity(400),
                         o2_shomate.heat_capacity(400))
        self.assertEqual(self.o2.standard_enthalpy(400),
                         o2_shomate.standard_enthalpy(400))
        self.assertEqual(self.o2.standard_entropy(400),
                         o2_shomate.standard_entropy(400))

    def test_temperature_range(self):
        self.assertEqual(self.o2.min_temperature, 100)
        self.assertEqual(self.o2.max_temperature, 6000)
        self.assertEqual(self.o2.temperature_range, (100, 6000))

    def test_zero_point(self):
        # Expected value is script of Prof. Kumagai
        self.assertAlmostEqual(Gas.O2.zero_point_vibrational_energy, 0.098/2, 2)
        self.assertAlmostEqual(Gas.N2.zero_point_vibrational_energy, 0.146/2, 2)

    def test_energy_shift(self):
        
        for t in range(0, 1001, 100):
            for p in [REFERENCE_PRESSURE/100, REFERENCE_PRESSURE]:
                if t == 0:
                    expected = self.o2.zero_point_vibrational_energy
                else:
                    expected = self.o2.zero_point_vibrational_energy + \
                               self.o2.vib_rot_term(t) + \
                               t * self.o2.r_ln_p_p0_ev_per_atom(p)

                actual = self.o2.energy_shift(temperature=t,
                                              pressure=p)
                self.assertAlmostEqual(expected, actual, 5)




# -*- coding: utf-8 -*-

import tempfile

from vise.input_set.prior_info import PriorInfo
from vise.util.testing import ViseTest


class PriorInfoTest(ViseTest):

    def setUp(self):
        """ """
        self.nonmagnetic_insulator = PriorInfo(energy_per_atom=-0.5,
                                               band_gap=1.2,
                                               total_magnetization=0.000001,
                                               data_source="Materials Project",
                                               is_cluster=False,
                                               mag_threshold=0.001,
                                               band_gap_threshold=0.1,
                                               incar={"NUPDOWN": 2})

    def test_dict(self):
        # round trip test of dict
        d = self.nonmagnetic_insulator.as_dict()
        prior_info_from_dict = PriorInfo.from_dict(d)
        self.assertTrue(d == prior_info_from_dict.as_dict())

    def test_yaml(self):
        """ round trip test of yaml"""
        tmp_file = tempfile.NamedTemporaryFile()
        self.nonmagnetic_insulator.dump_yaml(tmp_file.name)
        prior_info_from_json = PriorInfo.load_yaml(tmp_file.name)
        self.assertEqual(prior_info_from_json.as_dict(),
                         self.nonmagnetic_insulator.as_dict())

    def test_json(self):
        """ round trip test of to_json and from_json """
        tmp_file = tempfile.NamedTemporaryFile()
        self.nonmagnetic_insulator.dump_json(tmp_file.name)
        prior_info_from_json = PriorInfo.load_json(tmp_file.name)
        self.assertEqual(prior_info_from_json.as_dict(),
                         self.nonmagnetic_insulator.as_dict())

    def test_msonable(self):
        self.assertMSONable(self.nonmagnetic_insulator)



# -*- coding: utf-8 -*-

import json
import os
import re
import shutil
from glob import glob
from pathlib import Path
import tempfile
from typing import Optional, List
from uuid import uuid4

import numpy as np
from custodian.vasp.jobs import VaspJob
from monty.json import MSONable
from monty.json import MontyEncoder
from monty.serialization import loadfn
from pymatgen.core.structure import Structure
from pymatgen.io.vasp import Kpoints, Vasprun
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer

from vise.config import (
    SYMMETRY_TOLERANCE, ANGLE_TOL, KPT_INIT_DENSITY, KPT_FACTOR)
from vise.input_set.incar import Incar
from vise.input_set.input_set import ViseInputSet
from vise.input_set.task import Task, LATTICE_RELAX_TASK
from vise.input_set.xc import Xc
from vise.util.error_classes import VaspNotConvergedError, KptNotConvergedError
from vise.util.logger import get_logger

""" Provides structure optimization and kpt convergence jobs for VASP runs. """

logger = get_logger(__name__)


VASP_INPUT_FILES = {"INCAR", "POSCAR", "POTCAR", "KPOINTS"}
VASP_SAVED_FILES = {"INCAR", "vasprun.xml", "CONTCAR", "OUTCAR", "PROCAR"}
VASP_FINISHED_FILES = {i + ".finish" for i in VASP_SAVED_FILES}


def rm_wavecar(remove_current: bool,
               remove_subdirectories: Optional[bool] = False) -> None:
    """Remove WAVECARs at the current directory and subdirectories."""
    logger.info(f"Remove WAVECAR in current directory: {remove_current}")
    logger.info(f"Remove WAVECAR in   sub directories: {remove_subdirectories}")

    if remove_current:
        try:
            os.remove("WAVECAR")
        except FileNotFoundError:
            logger.log("WAVECAR does not exist at the current directory.")
            pass

    if remove_subdirectories:
        for wavecar in Path(".").glob('**/WAVECAR'):
            wavecar.unlink()


class StructureOptResult(MSONable):
    def __init__(self,
                 uuid: int,
                 energy_per_atom: float,
                 num_kpt: list,
                 final_structure: Structure,
                 final_sg: int,
                 kpt_density: float = None,
                 initial_structure: Optional[Structure] = None,
                 initial_sg: Optional[int] = None,
                 prev_structure_opt_uuid: Optional[int] = None):
        """Container object with structure optimization results.

        Args:
            uuid (int):
                Int of UUID of the present calculation.
            energy_per_atom (float):
                Energy per atom in eV.
            num_kpt (list):
                Numbers of k-points along three directions. E.g. [4, 6, 8]
            final_structure (Structure):
                Final structure.
            final_sg (int):
                Final space group.
            kpt_density (float):
                K-point density used for determining num_kpt.
            initial_structure (Structure):
                Initial structure.
            initial_sg (int):
                Initial space group.
            prev_structure_opt_uuid (int):
                Previous StructureOptResult object.
        """

        self.uuid = uuid
        self.energy_per_atom = energy_per_atom
        self.num_kpt = list(num_kpt)
        self.kpt_density = kpt_density
        self.final_structure = final_structure
        self.final_sg = final_sg
        self.initial_structure = initial_structure
        self.initial_sg = initial_sg
        self.prev_structure_opt_uuid = prev_structure_opt_uuid

    @classmethod
    def from_dir(cls,
                 dir_name: str,
                 kpoints: Optional[str] = "KPOINTS",
                 poscar: Optional[str] = "POSCAR.orig",
                 contcar: Optional[str] = "CONTCAR.finish",
                 vasprun: Optional[str] = "vasprun.xml.finish",
                 symprec: float = SYMMETRY_TOLERANCE,
                 angle_tolerance: float = ANGLE_TOL,
                 prev_structure_opt: Optional["StructureOptResult"] = None,
                 ) -> "StructureOptResult":
        """Constructor from directory and previous StructureOptResult if exists

        Generate initial_structure and initial_sg from previous
        StructureOptResult if exists.

        Args:
            dir_name (str):
                Dirname
            kpoints (str):
            poscar (str):
            contcar (str):
            vasprun (str):
                Vasp input and output file names
            symprec (float):
                Distance precision in Angstrom used for symmetry analysis.
            angle_tolerance (float):
                Angle precision in degrees used for symmetry analysis.
            prev_structure_opt (StructureOptResult):
                Previous StructureOptResult

        Return:
            StructureOptResult object
        """
        d_path = Path(dir_name)

        k = Kpoints.from_file(d_path / kpoints)
        num_kpt = k.kpts[0]

        try:
            vise = ViseInputSet.load_json(d_path / "vise.json")
            kpt_density = vise.kwargs["kpt_density"]
        except FileNotFoundError:
            print("vise.json does not exist")
            raise

        final_structure = Structure.from_file(d_path / contcar)
        sga = SpacegroupAnalyzer(structure=final_structure,
                                 symprec=symprec,
                                 angle_tolerance=angle_tolerance)
        symmetry_dataset = sga.get_symmetry_dataset()
        final_sg = symmetry_dataset["number"]

        v = Vasprun(d_path / vasprun)
        energy_per_atom = v.final_energy / len(final_structure)

        if prev_structure_opt:
            initial_structure = prev_structure_opt.final_structure
            initial_sg = prev_structure_opt.final_sg
            prev_structure_opt_uuid = prev_structure_opt.uuid
        else:
            initial_structure = Structure.from_file(d_path / poscar)
            sga = SpacegroupAnalyzer(structure=initial_structure,
                                     symprec=symprec,
                                     angle_tolerance=angle_tolerance)
            initial_symmetry_dataset = sga.get_symmetry_dataset()
            initial_sg = initial_symmetry_dataset["number"]
            prev_structure_opt_uuid = None

        return cls(uuid=int(uuid4()),
                   energy_per_atom=energy_per_atom,
                   num_kpt=num_kpt,
                   kpt_density=kpt_density,
                   final_structure=final_structure,
                   final_sg=final_sg,
                   initial_structure=initial_structure,
                   initial_sg=initial_sg,
                   prev_structure_opt_uuid=prev_structure_opt_uuid)

    @property
    def is_sg_changed(self):
        return self.initial_sg != self.final_sg

    @property
    def dirname(self):
        """Used for generating and parsing directory in k-point convergence"""

        name = [f"kpt{self.num_kpt[0]}x{self.num_kpt[1]}x{self.num_kpt[2]}",
                f"pre-sg{self.initial_sg}", f"pos-sg{self.final_sg}"]
        return '_'.join(name)

    def to_json_file(self, filename: str = "structure_opt.json") -> None:
        with open(filename, 'w') as fw:
            json.dump(self.as_dict(), fw, indent=2, cls=MontyEncoder)

    @classmethod
    def load_json(cls, filename: str = "structure_opt.json"):
        return loadfn(filename)

    def __str__(self):
        outs = [f"k-points: {self.num_kpt}",
                f"k-point density {self.kpt_density}",
                f"space group: {self.initial_sg} -> {self.final_sg}",
                f"energy per atom (eV): {self.energy_per_atom}",
                f"uuid: {self.uuid}",
                f"prev_uuid: {self.prev_structure_opt_uuid}"]
        return "\n".join(outs)


class KptConvResult(MSONable):

    def __init__(self,
                 str_opts: List[StructureOptResult],
                 convergence_criterion: float,
                 num_kpt_check: int,
                 symprec: float,
                 angle_tolerance: float):
        """Container object with k-point convergence results."

        structure_opt_results (list):
            List of StructureOptResult. The inner objects need to be linked
            to be consistent with prev_structure_opt_uuid.
        convergence_criterion:
            Convergence energy criterion as a function of number of k points in
            eV / atom.
        num_kpt_check:
            Number of K-point set that is used for checking the convergence.
            For example, when it is 2, we calculate
            num_kpt = [4, 4, 4], [6, 6, 6], [8, 8, 8]
            and check whether result with [4, 4, 4] is converged with
            the results of [6, 6, 6] and [8, 8, 8].
        symprec (float):
            Distance precision in Angstrom used for symmetry analysis.
            This is also checked if the lattice constants are converged.
        """
        self.str_opts = str_opts
        self.convergence_criterion = convergence_criterion
        self.num_kpt_check = num_kpt_check
        self.symprec = symprec
        self.angle_tolerance = angle_tolerance

    @classmethod
    def from_dirs(cls,
                  convergence_criterion: float,
                  num_kpt_check: int,
                  symprec: float,
                  angle_tolerance: float,
                  dirs: Optional[list] = None,
                  str_opt_filename: str = "structure_opt.json"
                  ) -> "KptConvResult":
        """Constructor from directories

        convergence_criterion:
        num_kpt_check:
        symprec:
            See docstrings of the constructor.
        dirs:
            Directory names to be parsed.
        structure_opt_filename (list):
            Json filename for StructureOptResult.
        symprec (float):

        Return:
            KptConvResult object.
        """
        dirs = dirs or glob("*kpt*/")
        str_opts = [StructureOptResult.load_json(Path(d, str_opt_filename))
                    for d in dirs]

        return cls(cls.sort_results(str_opts), convergence_criterion,
                   num_kpt_check, symprec, angle_tolerance)

    @staticmethod
    def sort_results(results: list) -> list:
        """Sort StructureOptResult list using prev_structure_opt_uuid """
        sorted_results = [r for r in results if not r.prev_structure_opt_uuid]
        if len(sorted_results) > 1:
            raise ValueError(f"The number of structure optimization "
                             f"{len(sorted_results)} is more than 1.")

        for i in range(len(results)):
            sorted_results += \
                [r for r in results
                 if r.prev_structure_opt_uuid == sorted_results[-1].uuid]
        return sorted_results

    @property
    def space_groups(self):
        return \
            [self.str_opts[0].initial_sg] + [i.final_sg for i in self.str_opts]

    @property
    def converged_result(self):
        """Return converged StructureOptResult.

        If not converged, return False.
        """
        if len(self.str_opts) <= self.num_kpt_check:
            return False

        # The target index is the num_kpt_check-th last.
        target_idx = -(self.num_kpt_check + 1)

        # Number of space groups is incremented from calc results by 1
        # so need to check target_idx - 1.
        if len(set(self.space_groups[(target_idx - 1):])) > 1:
            return False

        for i in range(self.num_kpt_check):
            target = self.str_opts[target_idx]
            comparator = self.str_opts[target_idx + i]
            energy_diff = target.energy_per_atom - comparator.energy_per_atom
            if abs(energy_diff) > self.convergence_criterion:
                logger.info("Energy is not converged, yet")
                return False

            # Check convergence of lattice matrix.
            if not np.allclose(a=target.final_structure.lattice.matrix,
                               b=comparator.final_structure.lattice.matrix,
                               rtol=0, atol=self.symprec):
                logger.info("Structure is not converged, yet")
                return False

        return self.str_opts[target_idx]

    @property
    def is_sg_changed(self):
        return len(set(self.space_groups)) > 1

    def __str__(self):
        # TODO: Add lattice info
        outs = \
            [f"Initial space group: {self.str_opts[0].initial_sg}",
                f"Convergence criterion (eV/atom): {self.convergence_criterion}",
             f"Symprec (A): {self.symprec}",
             f"Angle tolerance (deg): {self.angle_tolerance}", "",
             f"{'sg':>4} {'kpt':>10} {'energy/atom':>12}"
             f" {'relative energy/atom':>21}"]

        ref_energy = self.str_opts[-1].energy_per_atom
        for result in self.str_opts:
            relative_energy = result.energy_per_atom - ref_energy
            outs.append(f"{result.final_sg:>4} " 
                        f"{'x'.join(map(str, result.num_kpt)):>10} "
                        f"{result.energy_per_atom:12.5f}"
                        f"{relative_energy:21.5f}")

        outs.append("")
        return "\n".join(outs)

    def to_json_file(self, filename: str = "kpt_conv.json") -> None:
        with open(filename, 'w') as fw:
            json.dump(self.as_dict(), fw, indent=2, cls=MontyEncoder)

    @classmethod
    def load_json(cls, filename="kpt_conv.json"):
        return loadfn(filename)

    # TODO: show/hold warning in case the energy is increased when the symmetry
    #  is increased.


class ViseVaspJob(VaspJob):
    def __init__(self,
                 vasp_cmd: list,
                 output_file: str = "vasp.out",
                 stderr_file: str = "std_err.txt",
                 suffix: str = "",
                 final: bool = True,
                 backup: bool = True,
                 gamma_vasp_cmd: Optional[list] = None,
                 auto_continue: bool = False,
                 # add original args.
                 ncl_vasp_cmd: Optional[list] = None,
                 auto_gamma_vasp: Optional[list] = True,
                 auto_ncl_vasp: Optional[list] = True):

        """
        Override constructor to close some options as some options use Incar
        instead of VisaIncar.

        Args: See docstrings of VaspJob.
        """

        self.ncl_vasp_cmd = ncl_vasp_cmd
        self.auto_gamma_vasp = auto_gamma_vasp
        self.auto_ncl_vasp = auto_ncl_vasp

        # Should be fine for vasp.5.4.4
        if auto_gamma_vasp and gamma_vasp_cmd is None:
            gamma_vasp_cmd = vasp_cmd[:-1]
            gamma_vasp_cmd.append(vasp_cmd[-1].replace("std", "gam"))

        if auto_ncl_vasp:
            try:
                incar = Incar.from_file("INCAR")
                ncl_calc = any([incar.get("LNONCOLLINEAR", False),
                                incar.get("LSORBIT", False)])
                if ncl_calc:
                    if ncl_vasp_cmd is None:
                        ncl_vasp_cmd = vasp_cmd[:-1]
                        ncl_vasp_cmd.append(vasp_cmd[-1].replace("std", "ncl"))
                    vasp_cmd = ncl_vasp_cmd
            except FileNotFoundError:
                pass

        # Note that only list instance is accepted for vasp_cmd in VaspJob at
        # ver.2019.8.24.
        super().__init__(vasp_cmd=vasp_cmd,
                         output_file=output_file,
                         stderr_file=stderr_file,
                         suffix=suffix,
                         final=final,
                         backup=backup,
                         auto_npar=False,
                         auto_gamma=True,
                         settings_override=None,
                         gamma_vasp_cmd=gamma_vasp_cmd,
                         copy_magmom=False,
                         auto_continue=auto_continue)

    def postprocess(self):
        """
        """
        if self.suffix != "":
            for f in VASP_SAVED_FILES | {self.output_file}:
                if os.path.exists(f):
                    if f in ["PROCAR", "OUTCAR", "vasprun.xml"]:
                        shutil.move(f, f"{f}{self.suffix}")
                    else:
                        shutil.copy(f, f"{f}{self.suffix}")

        # Remove continue.json if a subsequent job is run in
        # the same directory, will not restart this job.
        if os.path.exists("continue.json"):
            os.remove("continue.json")

    @classmethod
    def structure_optimization_run(
            cls,
            vasp_cmd: list,
            prev_structure_opt: StructureOptResult = None,
            gamma_vasp_cmd: Optional[list] = None,
            max_relax_num: int = 10,
            removes_wavecar: bool = False,
            std_out: str = "vasp.out",
            move_unimportant_files: bool = True,
            left_files: Optional[list] = None,
            removed_files: Optional[list] = None,
            symprec: float = SYMMETRY_TOLERANCE,
            angle_tolerance: float = ANGLE_TOL
            ) -> None:
        """Vasp job for structure optimization

        Note1: The shell script name need to be finished with ".sh".

        Args:
            vasp_cmd (list):
                Vasp command.
            gamma_vasp_cmd: (str/list/None)
                Gamma version of vasp command.
            max_relax_num (int):
                Maximum number of vasp calculations.
            removes_wavecar (bool):
                Whether to remove the WAVECAR files after geometry optimization.
            std_out (str):
                Name of the file showing the standard output.
            move_unimportant_files (bool):
                Whether to move relatively unimportant results to calc_results.
            left_files (list):
                List of left files at the calculation directory.
            removed_files (list):
                List of removed files from files directory.
            prev_structure_opt:
            symprec:
            angle_tolerance:
                See docstrings of StructureOptResult.

        Yield:
            ViseVaspJob class object.
        """
        backup_initial_input_files = True
        for job_number in range(1, max_relax_num + 1):
            yield cls(vasp_cmd=vasp_cmd,
                      gamma_vasp_cmd=gamma_vasp_cmd,
                      final=True,
                      backup=backup_initial_input_files,
                      suffix=f".{job_number}")
            shutil.copy(".".join(["CONTCAR", str(job_number)]), "POSCAR")
            backup_initial_input_files = False

            if len(Vasprun(f"vasprun.xml.{job_number}").ionic_steps) == 1:
                break
        else:
            raise VaspNotConvergedError("Structure optimization not converged")

        left_files = set(left_files) if left_files else set()
        left_files.update(VASP_INPUT_FILES |
                          {"WAVECAR", "vise.json", "structure_opt.json"})

        for f in VASP_SAVED_FILES | {std_out}:
            finish_name = f"{f}.finish"
            shutil.move(f"{f}.{job_number}", finish_name)
            left_files.add(finish_name)

        removed_files = removed_files or []
        rm_wavecar(removes_wavecar)

        if move_unimportant_files:
            dir_name = "files"
            if os.path.exists(dir_name):
                logger.warning(f"{dir_name} is being removed.")
                # `tempfile.mktemp` Returns an absolute pathname of a file that
                # did not exist at the time the call is made. We pass
                # dir=os.path.dirname(dir_name) here to ensure we will move
                # to the same filesystem. Otherwise, shutil.copy2 will be used
                # internally and the problem remains.
                tmp = tempfile.mktemp(dir=os.path.dirname(dir_name))
                # Rename the dir.
                shutil.move(dir_name, tmp)
                # And delete it.
                shutil.rmtree(tmp)

            # At this point, even if tmp is still being deleted,
            # there is no name collision.
            os.makedirs(dir_name)

        result = \
            StructureOptResult.from_dir(dir_name=".",
                                        symprec=symprec,
                                        angle_tolerance=angle_tolerance,
                                        prev_structure_opt=prev_structure_opt)

        result.to_json_file("structure_opt.json")

        for f in glob("*"):
            if not os.path.isfile(f):
                continue  # continue if f is directory.
            elif os.stat(f).st_size == 0:
                os.remove(f)  # remove empty files
            elif move_unimportant_files \
                    and f not in left_files \
                    and not re.match(r".+\.sh$", f):
                if any([re.search(p, f) for p in removed_files]):
                    os.remove(f)  # remove empty files
                else:
                    shutil.move(f, "files")

    @classmethod
    def kpt_converge(cls,
                     vasp_cmd: list,
                     structure: Structure = None,
                     task: Task = Task.structure_opt,
                     xc: Xc = Xc.pbe,
                     user_incar_settings: Optional[dict] = None,
                     initial_kpt_density: Optional[float] = KPT_INIT_DENSITY,
                     gamma_vasp_cmd: Optional[list] = None,
                     max_relax_num: int = 10,
                     max_kpt_num: int = 10,
                     convergence_criterion: float = 0.003,
                     num_kpt_check: int = 2,
                     removes_wavecar: bool = False,
                     left_files: Optional[list] = None,
                     removed_files: Optional[list] = None,
                     std_out: str = "vasp.out",
                     symprec: float = SYMMETRY_TOLERANCE,
                     angle_tolerance: float = ANGLE_TOL,
                     **vis_kwargs) -> None:
        """"Vasp job for checking k-point convergence

        Args:
            structure (Structure):
                Input structure. If not set, parse POSCAR file.
            vasp_cmd:
            gamma_vasp_cmd:
            max_relax_num:
            removes_wavecar:
            removed_files:
            std_out:
                See docstrings of structure_optimization_run.
            -------
            task:
            xc:
            user_incar_settings:
            initial_kpt_density:
                See docstrings of ViseInputSet.
            -------
            num_kpt_check:
            convergence_criterion:
                See docstrings of KptConvResult.
            -------
            max_kpt_num (str):
                Max k-point iteration number
            left_files (list):
                List of left files at the calculation directory.
            symprec (float):
                Distance precision in Angstrom used for symmetry analysis.
                This is also checked if the lattice constants are converged.
            angle_tolerance (float):
                Angle precision in degrees used for symmetry analysis.
            vis_kwargs:
                Used for ViseInputSet keyword args.
        Return:
            None
        """
        if task not in LATTICE_RELAX_TASK:
            raise ValueError(f"Task: {task} is not in lattice relax set.")

        structure = structure or Structure.from_file("POSCAR")
        kpt_conv = KptConvResult.from_dirs(convergence_criterion, num_kpt_check,
                                           symprec, angle_tolerance)
        left_files = left_files or []
        if kpt_conv.str_opts:
            logger.info(f"{len(kpt_conv.str_opts)} structure optimization is"
                        f"parsed.")
            is_sg_changed = kpt_conv.str_opts[-1].is_sg_changed
        else:
            is_sg_changed = None

        vis_kwargs.update({"task": task,
                           "xc": xc,
                           "user_incar_settings": user_incar_settings,
                           "symprec": symprec,
                           "angle_tolerance": angle_tolerance})

        while not kpt_conv.converged_result \
                and len(kpt_conv.str_opts) < max_kpt_num:

            prev_str_opt = kpt_conv.str_opts[-1] if kpt_conv.str_opts else None

            if is_sg_changed is False:
                prev_kpt = prev_str_opt.num_kpt
                kpt_density = prev_str_opt.kpt_density

                # Iteratively increase kpt_density until num_kpt is incremented.
                while True:
                    kpt_density *= KPT_FACTOR
                    vis = ViseInputSet.from_prev_calc(
                        dirname=prev_str_opt.dirname,
                        parse_calc_results=False,
                        parse_incar=True,
                        sort_structure=False,
                        standardize_structure=True,
                        files_to_transfer={"WAVECAR": "m"},
                        contcar_filename="CONTCAR.finish",
                        kpt_density=kpt_density,
                        **vis_kwargs)
                    # ex kpoints.kpts = [[5, 5, 4]], so we need [0].
                    kpt = vis.kpoints.kpts[0]
                    if (not kpt == prev_kpt
                            and all([i >= j for i, j in zip(kpt, prev_kpt)])):
                        break
            else:
                if is_sg_changed is True:
                    logger.warning("Space group is changed during the kpoint "
                                   "convergence, So the kpoint check is "
                                   "reiterated.")
                    name = "/".join([prev_str_opt.dirname, "CONTCAR.finish"])
                    structure = Structure.from_file(name)
                # When symmetry is changed, kpt convergence is tested from
                # the scratch.
                kpt_density = initial_kpt_density

                vis = ViseInputSet.make_input(
                    structure=structure,
                    kpt_density=kpt_density,
                    **vis_kwargs)

            vis.write_input(".")
            # Need to put forward the generator in structure_optimization_run
            # When symmetry is changed, custodian corrections are not inherited.
            for run in cls.structure_optimization_run(
                    vasp_cmd=vasp_cmd,
                    gamma_vasp_cmd=gamma_vasp_cmd,
                    max_relax_num=max_relax_num,
                    std_out=std_out,
                    left_files=left_files,
                    removed_files=removed_files,
                    symprec=symprec,
                    angle_tolerance=angle_tolerance,
                    prev_structure_opt=prev_str_opt):
                yield run

            str_opt = StructureOptResult.load_json("structure_opt.json")
            kpt_conv.str_opts.append(str_opt)
            os.mkdir(str_opt.dirname)
            for f in glob("*"):
                if ((f == "files" or os.path.isfile(f))
                        and not re.match(r".+\.sh$", f)
                        and not f == "WAVECAR"
                        and f not in left_files):
                    shutil.move(f, str_opt.dirname)

            is_sg_changed = str_opt.is_sg_changed

        rm_wavecar(remove_current=removes_wavecar, remove_subdirectories=True)

        if kpt_conv.converged_result:
            conv_dirname = Path(kpt_conv.converged_result.dirname)
            for f in VASP_INPUT_FILES | VASP_FINISHED_FILES - {"INCAR"}:
                os.symlink(str(conv_dirname / f), f)
            kpt_conv.to_json_file("kpt_conv.json")
        else:
            raise KptNotConvergedError(
                "Energy was not converged as function of k-points numbers.")





# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import os
from copy import deepcopy

from argparse import Namespace
from shutil import rmtree
from unittest.mock import patch, Mock, MagicMock

from pymatgen.core.structure import IStructure

from vise.util.testing import ViseTest
from vise.cli.main_function import (
    vasp_settings_from_args, get_poscar_from_mp, vasp_set, vasp_run,
    chempotdiag, plot_band, plot_dos, band_gap)
from vise.cli.tests.test_main import default_vasp_args, symprec_args
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.util.logger import get_logger
from vise.config import KPT_DENSITY


logger = get_logger(__name__)


vasp_args = default_vasp_args


class VaspSettingsFromArgsTest(ViseTest):
    def setUp(self) -> None:
        vasp = {"ldauu": ["Mn", "5", "O", "3"],
                "ldaul": ["Mn", "3", "O", "2"],
                "vise_opts": ["only_even", "True"],
                "user_incar_settings": ["POTIM", "0.4"],
                "additional_user_incar_settings": ["ALGO", "Fast"],
                "potcar_set": ["Mn_pv"],
                "potcar_set_name": "normal",
                "charge": 1}
        self.args = Namespace(**vasp)

    def test(self):
        user_incar_settings, vis_base_kwargs = \
            vasp_settings_from_args(self.args)

        expected = {"POTIM": 0.4, "ALGO": "Fast"}
        self.assertEqual(expected, user_incar_settings)

        expected = {"ldauu": {"Mn": 5, "O": 3},
                    "ldaul": {"Mn": 3, "O": 2},
                    "only_even": True,
                    "override_potcar_set": {"Mn": "Mn_pv"},
                    "potcar_set_name": "normal",
                    "charge": 1}
        self.assertEqual(expected, vis_base_kwargs)


class GetPoscarFromMpTest(ViseTest):
    def setUp(self) -> None:
        self.args_num = Namespace(number=110, poscar="POSCAR")
        self.kwargs = {"elements": ["Mg", "O"],
                       "e_above_hull": 1.5,
                       "molecules": True}
        self.args_elements = Namespace(**self.kwargs)
        self.args_none = Namespace()

    def test_number(self):
        get_poscar_from_mp(self.args_num)
        expected = IStructure.from_str("""Mg1
1.0
2.922478 0.000000 -1.033252
-1.461239 2.530940 -1.033252
0.000000 0.000000 3.099756
Mg
1
direct
0.000000 0.000000 0.000000 Mg""", fmt="poscar")
        actual = IStructure.from_file("POSCAR")
        self.assertEqual(expected, actual)

    @patch('vise.cli.main_function.make_poscars_from_mp')
    def test_mp_poscars(self, mock_make_poscars):
        get_poscar_from_mp(self.args_elements)
        mock_make_poscars.assert_called_with(**self.kwargs)

    def test_warning(self):
        expected = ['WARNING:vise.cli.main_function:Set mp number or elements']
        with self.assertLogs(level='WARNING') as cm:
            get_poscar_from_mp(self.args_none)
            self.assertEqual(expected, cm.output)

    def tearDown(self) -> None:
        if os.path.exists("POSCAR"):
            os.remove("POSCAR")


class VaspSetTest(ViseTest):
    def setUp(self) -> None:
        self.min_default_kwargs = {"poscar": "POSCAR",
                                   "kpt_density": KPT_DENSITY,
                                   "standardize_structure": True,
                                   "prior_info": True,
                                   }
        self.args_print = Namespace(json="vasp_input_set.json",
                                    **self.min_default_kwargs,
                                    **vasp_args, **symprec_args)
        self.prev_dir = "a"
        self.args_prev = Namespace(json=None,
                                   prev_dir=self.prev_dir,
                                   **self.min_default_kwargs,
                                   **vasp_args, **symprec_args)

        self.args_normal = Namespace(json=None,
                                     **self.min_default_kwargs,
                                     dirs=["."],
                                     prev_dir=None,
                                     **vasp_args, **symprec_args)

    @patch('vise.cli.main_function.ViseInputSet.load_json')
    def test_print(self, mock):
        vasp_set(self.args_print)
        mock.assert_called_with("vasp_input_set.json")

    @patch('vise.cli.main_function.ViseInputSet.from_prev_calc')
    def test_prev_dir(self, mock):
        vasp_set(self.args_prev)
        kwargs = {"override_potcar_set": vasp_args["potcar_set"] or {},
                  "potcar_set_name": vasp_args["potcar_set_name"],
                  "ldauu": vasp_args["ldauu"] or {},
                  "ldaul": vasp_args["ldaul"] or {},
                  "charge": vasp_args["charge"],
                  "symprec": symprec_args["symprec"],
                  "angle_tolerance": symprec_args["angle_tolerance"],
                  "kpt_density": self.min_default_kwargs["kpt_density"],
                  "standardize_structure":
                      self.min_default_kwargs["standardize_structure"],
                  "files_to_transfer":
                      {"CHGCAR": "C", "WAVECAR": "M", "WAVEDER": "M"},
                  "task": Task.from_string(vasp_args["task"]),
                  "xc": Xc.from_string(vasp_args["xc"])}

        mock.assert_called_with("a", **kwargs)

    @patch('vise.cli.main_function.Structure.from_file')
    @patch('vise.cli.main_function.ViseInputSet.make_input')
    def test_input_set(self, mock, mock_structure):
        vasp_set(self.args_normal)

        user_incar_settings = {}

        kwargs = {"structure": mock_structure(),
                  "override_potcar_set": vasp_args["potcar_set"] or {},
                  "potcar_set_name": vasp_args["potcar_set_name"],
                  "ldauu": vasp_args["ldauu"] or {},
                  "ldaul": vasp_args["ldaul"] or {},
                  "charge": vasp_args["charge"],
                  "symprec": symprec_args["symprec"],
                  "angle_tolerance": symprec_args["angle_tolerance"],
                  "kpt_density": self.min_default_kwargs["kpt_density"],
                  "standardize_structure":
                      self.min_default_kwargs["standardize_structure"],
                  "task": Task.from_string(vasp_args["task"]),
                  "xc": Xc.from_string(vasp_args["xc"]),
                  "user_incar_settings": user_incar_settings}

        mock.assert_called_with(**kwargs)


class VaspRunTest(ViseTest):
    def setUp(self) -> None:
        self.min_default_kwargs = {"poscar": "POSCAR",
                                   "kpt_density": KPT_DENSITY,
                                   "standardize_structure": True,
                                   "prior_info": True,
                                   }
        self.args_print = Namespace(json="vasp_input_set.json",
                                    **self.min_default_kwargs,
                                    **vasp_args, **symprec_args)


class TestChemPotDiag(ViseTest):
    def setUp(self) -> None:
        self.kwargs_1 = {
            "draw_phase_diagram": True,
            "vasp_dirs": None,
            "vasprun": None,
            "elements": ["Mg", "O"],
            "target_comp": "MgO",
            "filename": None,
            "parse_gas": True,
            "partial_pressures": ["O", "1e+5"],
            "temperature": 1000.0}

        self.from_mp_pd_filename = Namespace(**self.kwargs_1)

        self.kwargs_2 = {
            "draw_phase_diagram": False,
            "vasp_dirs": ["a_dir", "b_dir"],
            "vasprun": "vasprun.xml.finish",
            "elements": None,
            "target_comp": "MgO",
            "filename": "cpd.pdf",
            "parse_gas": False,
            "partial_pressures": ["O", "1e+5"],
            "temperature": 0.0}

        self.from_dir_cpd = Namespace(**self.kwargs_2)

    @patch('vise.cli.main_function.PDPlotter')
    @patch('vise.cli.main_function.PhaseDiagram')
    @patch('vise.cli.main_function.FreeEnergyEntrySet.from_mp')
    def test_from_mp_pd_filename(self, mock_entry_set, mock_pd, mock_pd_plot):
        chempotdiag(self.from_mp_pd_filename)
        mock_entry_set.assert_called_with(self.kwargs_1["elements"])
        mock_pd.assert_called_with(entries=mock_entry_set().entries)
        mock_pd_plot.assert_called_with(mock_pd())

    @patch('vise.cli.main_function.ChemPotDiag.from_phase_diagram')
    @patch('vise.cli.main_function.PhaseDiagram')
    @patch('vise.cli.main_function.FreeEnergyEntrySet.from_vasp_files')
    def test_from_dir_cpd(self, mock_entry_set, mock_pd, mock_cpd):
        chempotdiag(self.from_dir_cpd)
        mock_entry_set.assert_called_with(
            directory_paths=self.kwargs_2["vasp_dirs"],
            vasprun=self.kwargs_2["vasprun"],
            parse_gas=self.kwargs_2["parse_gas"],
            temperature=self.kwargs_2["temperature"],
            partial_pressures=self.kwargs_2["partial_pressures"])
        mock_pd.assert_called_with(entries=mock_entry_set().entries)
        mock_cpd.assert_called_with(mock_pd(),
                                    target_comp=self.kwargs_2["target_comp"])


class TestPlotBand(ViseTest):
    def setUp(self) -> None:
        self.kwargs = {"kpoints": "KPOINTS",
                       "vasprun": "1/vasprun.xml",
                       "vasprun2": "2/vasprun.xml",
                       "absolute": True,
                       "y_range": False,
                       "legend": False}
        self.args = Namespace(filename="band.pdf", **self.kwargs, **symprec_args)

    @patch('vise.cli.main_function.PrettyBSPlotter.from_vasp_files')
    def test_get_band_plot(self, mock):
        plot_band(self.args)
        kwargs = deepcopy(self.kwargs)
        kwargs["kpoints_filenames"] = kwargs.pop("kpoints")
        kwargs["vasprun_filenames"] = kwargs.pop("vasprun")
        kwargs["vasprun2_filenames"] = kwargs.pop("vasprun2")
        mock.assert_called_with(**kwargs, **symprec_args)


class TestGetDosPlotter(ViseTest):
    def setUp(self) -> None:
        self.kwargs = {"vasprun": "vasprun.xml",
                       "cbm_vbm": [3, 1],
                       "pdos_type": "element",
                       "specific": None,
                       "orbital": True,
                       "x_range": False,
                       "ymaxs": [2.0, 3.0],
                       "absolute": True,
                       "legend": False,
                       "crop_first_value": True}
        self.args = Namespace(filename="dos.pdf", **self.kwargs, **symprec_args)

    @patch('vise.cli.main_function.get_dos_plot')
    def test_get_dos_plot(self, mock):
        plot_dos(self.args)
        kwargs = deepcopy(self.kwargs)
        kwargs["zero_at_efermi"] = not kwargs.pop("absolute")
        kwargs["xlim"] = kwargs.pop("x_range")
        mock.assert_called_with(**kwargs, **symprec_args)


class TestBandGap(ViseTest):
    def setUp(self) -> None:
        self.kwargs = {"vasprun": "vasprun.xml", "outcar": "OUTCAR"}
        self.args = Namespace(**self.kwargs)

    @patch('vise.cli.main_function.band_gap_properties',
           return_value=({"a": 1}, {"b": 2}, {"c": 3}))
    def test_succeed(self, mock):
        band_gap(self.args)
        mock.assert_called_with(**self.kwargs)




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from argparse import Namespace
import os
import shutil
from pathlib import Path

from vise.util.testing import ViseTest
from vise.custodian_extension.jobs import ViseVaspJob
from vise.cli.main import (
    simple_override, parse_args)
from vise.cli.main_tools import get_default_args
from vise.util.mp_tools import make_poscars_from_mp
from vise.input_set.input_set import ViseInputSet
from vise.config import (
    SYMMETRY_TOLERANCE, ANGLE_TOL, KPT_DENSITY, KPT_INIT_DENSITY, TIMEOUT)


parent_dir = Path(__file__).parent


class SimpleOverrideTest(ViseTest):
    def setUp(self) -> None:
        vise_yaml = parent_dir / ".." / ".." / "test_files" / "vise.yaml"
        shutil.copy(str(vise_yaml), "vise.yaml")

    def test(self):
        d = {"xc": "pbesol", "symprec": 1}
        simple_override(d, ["xc"])
        expected = {'xc': 'hse', 'symprec': 1}

        self.assertEqual(expected, d)

    def tearDown(self) -> None:
        os.remove("vise.yaml")


vis_defaults = get_default_args(ViseInputSet.make_input)
vis_defaults.update(ViseInputSet.TASK_OPTIONS)
vis_defaults.update(ViseInputSet.XC_OPTIONS)

kpt_conv_defaults = get_default_args(ViseVaspJob.kpt_converge)

# func is a pointer so need to point the same address.
# The directly passed args from argpaser.
default_vasp_args = {
    "potcar_set": None,
    "potcar_set_name": vis_defaults["potcar_set_name"],
    "xc": vis_defaults["xc"],
    "task": vis_defaults["task"],
    "vise_opts": None,
    "user_incar_settings": None,
    "additional_user_incar_settings": None,
    "ldauu": vis_defaults["ldauu"],
    "ldaul": vis_defaults["ldaul"],
    "charge": None}

symprec_args = {"symprec": SYMMETRY_TOLERANCE,  "angle_tolerance": ANGLE_TOL}


class MainGetPoscarsTest(ViseTest):

    def test_get_poscars_wo_options(self):
        actual = parse_args(["gp"])
        print(actual)
        d = get_default_args(make_poscars_from_mp)
        # func is a pointer so need to point the same address.
        expected = Namespace(
            poscar=None,
            number=None,
            elements=None,
            e_above_hull=d["e_above_hull"],
            molecules=d["molecules"],
            func=actual.func)
#        self.assertEqual(expected, actual)

    def test_get_poscars_w_options(self):
        parsed_args = parse_args(["gp",
                                  "-p", "POSCAR",
                                  "-n", "123",
                                  "-e", "Mg", "O",
                                  "--e_above_hull", "0.1",
                                  "--molecules", "F"])

        expected = Namespace(
            poscar="POSCAR",
            number=123,
            elements=["Mg", "O"],
            e_above_hull=0.1,
            molecules=False,
            func=parsed_args.func)
        self.assertEqual(expected, parsed_args)


class MainVaspSetTest(ViseTest):
    def test_vasp_set_wo_options(self):
        parsed_args = parse_args(["vs"])
        # func is a pointer so need to point the same address.
        expected = Namespace(
            json=None,
            poscar="POSCAR",
            kpt_density=KPT_DENSITY,
            standardize_structure=True,
            prior_info=True,
            dirs=None,
            prev_dir=None,
            func=parsed_args.func,
            **default_vasp_args, **symprec_args)
        self.assertEqual(expected, parsed_args)

    def test_vasp_set_yaml(self):
        vise_yaml = parent_dir / ".." / ".." / "test_files" / "vise.yaml"
        shutil.copy(str(vise_yaml), str("vise.yaml"))
        parsed_args = parse_args(["vs",
                                  "--ldauu", "Mg", "5",
                                  "-d", "c"])
        os.remove("vise.yaml")
        expected = Namespace(
            potcar_set=default_vasp_args["potcar_set"],
            potcar_set_name="gw",
            xc="hse",
            task=default_vasp_args["task"],
            vise_opts=default_vasp_args["vise_opts"],
            user_incar_settings=default_vasp_args["user_incar_settings"],
            additional_user_incar_settings=
            default_vasp_args["additional_user_incar_settings"],
            ldauu=["Mg", "5"],
            ldaul=["Mn", "2", "O", "1"],
            charge=default_vasp_args["charge"],
            symprec=2,
            angle_tolerance=symprec_args["angle_tolerance"],
            json=None,
            poscar="POSCAR",
            kpt_density=KPT_DENSITY,
            standardize_structure=True,
            prior_info=True,
            dirs=None,
            prev_dir="c",
            func=parsed_args.func)

        self.assertEqual(expected, parsed_args)

    # Here, test vasp_parser and symprec_parser as well.
    def test_vasp_set_w_options(self):
        parsed_args = parse_args(["vs",
                                  "--potcar", "Mg_pv", "O_h",
                                  "--potcar_set_name", "gw",
                                  "-x", "pbesol",
                                  "-t", "band",
                                  "--vise_opts", "encut", "800",
                                  "--user_incar_settings", "LREAD", "F",
                                  "-auis", "ALGO", "D",
                                  "--ldauu", "Mg", "5",
                                  "--ldaul", "Mg", "1",
                                  "-c", "10",
                                  "--symprec", "0.0001",
                                  "--angle_tolerance", "10",
                                  "--json", "tmp.json",
                                  "--poscar", "POSCAR-tmp",
                                  "-k", "4.2",
                                  "-s", "F",
                                  "-pi", "T",
                                  "--dirs", "a", "b",
                                  "-d", "c"])

        expected = Namespace(
            potcar_set=["Mg_pv", "O_h"],
            potcar_set_name="gw",
            xc="pbesol",
            task="band",
            vise_opts=["encut", "800"],
            user_incar_settings=["LREAD", "F"],
            additional_user_incar_settings=["ALGO", "D"],
            ldauu=["Mg", "5"],
            ldaul=["Mg", "1"],
            charge=10.0,
            symprec=0.0001,
            angle_tolerance=10.0,
            json="tmp.json",
            poscar="POSCAR-tmp",
            kpt_density=4.2,
            standardize_structure=False,
            prior_info=True,
            dirs=["a", "b"],
            prev_dir="c",
            func=parsed_args.func)

        self.assertEqual(expected, parsed_args)


class MainVaspRunTest(ViseTest):
    def test_vasp_run_wo_options(self):
        parsed_args = parse_args(["vr", "-v", "vasp", "cmd"])
        # func is a pointer so need to point the same address.
        expected = Namespace(
            print=False,
            json_file="str_opt.json",
            vasp_cmd=["vasp", "cmd"],
            handler_name="default",
            timeout=TIMEOUT,
            remove_wavecar=False,
            max_relax_num=kpt_conv_defaults["max_relax_num"],
            left_files=kpt_conv_defaults["left_files"],
            func=parsed_args.func,
            **symprec_args)
        self.assertEqual(expected, parsed_args)

    def test_vasp_set_w_options(self):
        parsed_args = parse_args(["vr",
                                  "--print",
                                  "--json_file", "test.json",
                                  "-v", "vasp", "cmd",
                                  "--handler_name", "dielectric",
                                  "--timeout", "1000",
                                  "--remove_wavecar",
                                  "--max_relax_num", "10",
                                  "--left_files", "POSCAR", "PCDAT",
                                  ])
        expected = Namespace(
            print=True,
            json_file="test.json",
            vasp_cmd=["vasp", "cmd"],
            handler_name="dielectric",
            timeout=1000,
            remove_wavecar=True,
            max_relax_num=10,
            left_files=["POSCAR", "PCDAT"],
            func=parsed_args.func,
            **symprec_args)
        self.assertEqual(expected, parsed_args)


class MainKptConvTest(ViseTest):
    def test_kpt_conv_wo_options(self):
        parsed_args = parse_args(["kc", "-v", "vasp", "cmd"])
        # func is a pointer so need to point the same address.
        expected = Namespace(
            print=False,
            json_file="kpt_conv.json",
            vasp_cmd=["vasp", "cmd"],
            initial_kpt_density=KPT_INIT_DENSITY,
            handler_name="default",
            timeout=TIMEOUT,
            remove_wavecar=False,
            max_relax_num=kpt_conv_defaults["max_relax_num"],
            convergence_criterion=kpt_conv_defaults["convergence_criterion"],
            left_files=kpt_conv_defaults["left_files"],
            func=parsed_args.func,
            **default_vasp_args, **symprec_args)
        self.assertEqual(expected, parsed_args)

    def test_kpt_conv_w_options(self):
        parsed_args = parse_args(["kc",
                                  "--json_file", "test.json",
                                  "-v", "vasp", "cmd",
                                  "-ikd", "3.5",
                                  "--handler_name", "dielectric",
                                  "--timeout", "1000",
                                  "--remove_wavecar",
                                  "--max_relax_num", "10",
                                  "--criterion", "0.01",
                                  "--left_files", "POSCAR", "PCDAT"])
        expected = Namespace(
            print=False,
            json_file="test.json",
            vasp_cmd=["vasp", "cmd"],
            initial_kpt_density=3.5,
            handler_name="dielectric",
            timeout=1000,
            remove_wavecar=True,
            max_relax_num=10,
            convergence_criterion=0.01,
            left_files=["POSCAR", "PCDAT"],
            func=parsed_args.func,
            **default_vasp_args, **symprec_args)
        self.assertEqual(expected, parsed_args)


class MainChemPotDiagTest(ViseTest):
    def test_vasp_run_wo_options(self):
        parsed_args = parse_args(["cpd"])
        expected = Namespace(
            draw_phase_diagram=False,
            vasp_dirs=None,
            vasprun="vasprun.xml",
            elements=None,
            target_comp=None,
            filename=None,
            parse_gas=True,
            partial_pressures=None,
            temperature=0.0,
            func=parsed_args.func)
        self.assertEqual(expected, parsed_args)

    def test_vasp_run_w_options(self):
        parsed_args = parse_args(["cpd",
                                  "-dpd",
                                  "-d", "a_dir", "b_dir",
                                  "-v", "vasprun.xml.finish",
                                  "-e", "Mg", "O",
                                  "-c", "MgO",
                                  "-f", "cpd.pdf",
                                  "-pg", "F",
                                  "-pp", "O", "1e+5",
                                  "-t", "1000"])

        expected = Namespace(
            draw_phase_diagram=True,
            vasp_dirs=["a_dir", "b_dir"],
            vasprun="vasprun.xml.finish",
            elements=["Mg", "O"],
            target_comp="MgO",
            filename="cpd.pdf",
            parse_gas=False,
            partial_pressures=["O", "1e+5"],
            temperature=1000.0,
            func=parsed_args.func)
        self.assertEqual(expected, parsed_args)


# class MainPlotBandTest(ViseTest):
#     def



# !/usr/bin/env python
# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import argparse
import sys


def band_gap(args) -> None:
    print(args)
    try:
        print(f"band gap info")
    except TypeError:
        print("Metallic system")


def parse_args(args):

    parser = argparse.ArgumentParser(
        description="""                            
    Vise is a package that helps researchers to do first-principles calculations 
    with the VASP code.""",
        epilog=f"""                                 
    Author: Yu Kumagai""",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    subparsers = parser.add_subparsers()

    print("parser")
    print(parser)

    # -- band_gap --------------------------------------------------------------
    parser_band_gap = subparsers.add_parser(
        name="band_gap",
        description="Calculate the band gap from vasprun.xml",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        aliases=['bg'])

    parser_band_gap.add_argument(
        "-v", dest="vasprun", type=str, default="vasprun.xml", metavar="FILE")
    parser_band_gap.add_argument(
        "-o", dest="outcar", type=str, default="OUTCAR", metavar="FILE")
    parser_band_gap.set_defaults(func=band_gap)
    # args = parser.parse_args()
    # args.func(args)
    return parser.parse_args(args)


def main():
    args = parse_args(sys.argv[1:])
    args.func(args)


if __name__ == "__main__":
    main()
#    parse_args()



# -*- coding: utf-8 -*-

from collections import OrderedDict, defaultdict
from typing import Dict, List

import numpy as np

from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.dos import Dos
from pymatgen.electronic_structure.dos import add_densities
from pymatgen.electronic_structure.plotter import DosPlotter
from pymatgen.io.vasp import Vasprun
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.util.string import latexify_spacegroup, latexify

from vise.config import SYMMETRY_TOLERANCE, ANGLE_TOL
from vise.util.logger import get_logger
from vise.util.error_classes import InvalidStructureError


logger = get_logger(__name__)


class ViseDosPlotter(DosPlotter):

    def get_plot(self,
                 xlim: List[float] = None,
                 ylims: List[List[float]] = None,
                 cbm_vbm: List[float] = None,
                 legend: bool = True,
                 crop_first_value: bool = False,
                 title: str = None):
        """Get a matplotlib pyplot for the density of states.

        Override get_plot of DosPlotter.

        Args:
            xlim (list):
                Specifies the x-axis limits. Set to None for automatic
                determination.
            ylims (list):
                Specifies the y-axes limits. Two types of input.
                [[y1min, y1max], [y2min, y2max], ..]
            cbm_vbm (list):
                Specify cbm and vbm [cbm, vbm]
            legend (bool):
                Whether to show the figure legend.
            crop_first_value (bool):
                Whether to crop the fist DOS.
            title (str):
                Title of the figure

        Returns:
            Matplotlib pyplot
        """

        ncolors = max(3, len(self._doses))
        ncolors = min(9, ncolors)
        import palettable
        colors = palettable.colorbrewer.qualitative.Set1_9.mpl_colors

        y = None
        all_densities = []
        all_energies = []

        # The DOS calculated using VASP holds a spuriously large value at the
        # first mesh to keep the consistency with the integrated DOS in DOSCAR
        # file. An example is shown below.
        # ------------- DOSCAR --------------------
        # 10 10 1 0
        # 0.1173120E+02 0.5496895E-09 0.5496895E-09 0.5496895E-09 0.5000000E-15
        # 1.000000000000000E-004
        # CAR
        # unknown system
        #     23.00000000 - 9.00000000 3201 6.62000004 1.00000000
        #     -9.000 0.6000E+03 0.6000E+03 0.6000E+01 0.6000E+01 <-- large DOS
        #     -8.990 0.0000E+00 0.0000E+00 0.6000E+01 0.6000E+01

        i = 1 if crop_first_value else 0

        for key, dos in self._doses.items():
            energies = dos['energies'][i:]
            densities = {Spin(k): v[i:] for k, v in dos['densities'].items()}
            if not y:
                y = {Spin.up: np.zeros(energies.shape),
                     Spin.down: np.zeros(energies.shape)}
            all_energies.append(energies)
            all_densities.append(densities)

        # Make groups to be shown in the same figure.
        # Example, ZrTiSe4
        # keys = ['Total', 'Site:1 Zr-s', 'Site:1 Zr-p', 'Site:1 Zr-d',
        #         'Site:2 Ti-s', 'Site:2 Ti-p', 'Site:2 Ti-d', 'Site:3 Se-s',
        #         'Site:3 Se-p', 'Site:3 Se-d', 'Site:5 Se-s', 'Site:5 Se-p',
        #         'Site:5 Se-d']
        keys = list(self._doses.keys())
        grouped_keys = OrderedDict()
        for k in keys:
            first_word = k.split()[0]
            if first_word in grouped_keys:
                grouped_keys[first_word].append(k)
            else:
                grouped_keys[first_word] = [k]

        import matplotlib.pyplot as plt
        num_figs = len(grouped_keys)
        fig, axs = plt.subplots(num_figs, 1, sharex=True)

        if xlim:
            axs[0].set_xlim(xlim)

        n = 0
        for i, gk in enumerate(grouped_keys):
            all_pts = []
            for j, key in enumerate(grouped_keys[gk]):
                x = []
                y = []

                for spin in [Spin.up, Spin.down]:
                    if spin in all_densities[n]:
                        densities = list(int(spin) * all_densities[n][spin])
                        energies = list(all_energies[n])
                        x.extend(energies)
                        y.extend(densities)

                all_pts.extend(list(zip(x, y)))
                axs[i].plot(x, y, color=colors[j % ncolors], label=str(key),
                            linewidth=2)
                n += 1

            # plot vertical lines for band edges or Fermi level
            if self.zero_at_efermi:
                # plot a line
                axs[i].axvline(0, color="black", linestyle="--", linewidth=0.5)
                if cbm_vbm:
                    axs[i].axvline(cbm_vbm[0] - cbm_vbm[1], color="black",
                                   linestyle="--", linewidth=0.5)
            else:
                axs[i].axvline(self._doses[key]['efermi'],
                               color="black", linestyle="--", linewidth=0.5)
                if cbm_vbm:
                    axs[i].axvline(cbm_vbm[0], color="black", linestyle="--",
                                   linewidth=0.5)

            if legend:
                axs[i].legend(loc="best", markerscale=0.1)
                #                axs[i].legend(bbox_to_anchor=(1.1, 0.8), loc="best")
                leg = axs[i].get_legend()
                for legobj in leg.legendHandles:
                    legobj.set_linewidth(1.2)
                ltext = leg.get_texts()
                plt.setp(ltext, fontsize=7)

            axs[i].axhline(0, color="black", linewidth=0.5)

        if ylims and len(ylims) not in (num_figs, 2):
            raise ValueError("The number of y-ranges is not proper.")

        if ylims and len(ylims) == 2:
            if ylims[0][1] > 1.0:
                axs[0].set_ylim(ylims[0])
            for i in range(1, len(axs)):
                axs[i].set_ylim(ylims[1])

        elif ylims:
            for i in range(len(axs)):
                axs[i].set_ylim(ylims[i])
        # else:
        #     for i in range(len(axs)):
        #         ylim = axs[i].get_ylim()
        #         print(ylim)
        #         relevanty = [p[1] for p in all_pts
        #                      if ylim[0] < p[0] < ylim[1]]
        #         axs[i].set_ylim((min(relevanty), max(relevanty)))e

        axs[-1].set_xlabel('Energy (eV)')
        axs[0].set_ylabel("Total DOS (1/eV)")
        for i in range(1, len(axs)):
            axs[i].set_ylabel("DOS (1/eV)")

        plt.tight_layout()
        plt.subplots_adjust(left=None, bottom=None, right=None, top=None,
                            wspace=0, hspace=0.1)

        if title:
            axs[0].title.set_text(title)

        return plt


def get_dos_plot(vasprun: str,
                 cbm_vbm: list = None,
                 pdos_type: str = "element",
                 specific: list = None,
                 orbital: bool = True,
                 xlim: list = None,
                 ymaxs: list = None,
                 zero_at_efermi: bool = True,
                 legend: bool = True,
                 crop_first_value: bool = True,
                 show_spg: bool = True,
                 symprec: float = SYMMETRY_TOLERANCE,
                 angle_tolerance: float = ANGLE_TOL):
    """

    Args:
        vasprun (str):
            vasprun.xml-type file name
        cbm_vbm (list):
            List of [cbm, vbm]. This is used when band edge is determined from
            the band structure calculation.
        pdos_type (str): Plot type of PDOS.
            "element": PDOS grouped by element type
            "site": PDOS grouped by equivalent sites
            "none": PDOS are not grouped.
        specific (list): Show specific PDOS. If list elements are integers,
            PDOS at particular sites are shown. If elements are shown,
            PDOS of particular elements are shown.
            ["1", "2"] --> At site 1 and 2 compatible with pdos_type = "none"
            ["Mg", "O"] --> Summed at Mg and O sites compatible with
                            pdos_type = "element"
        orbital (bool):
            Whether to show orbital decomposed PDOS.
        xlim (list):
            Specifies x-axis limits.
            Set to None for automatic determination.
        ymaxs (list):
            Specifies the maxima of absolute y-axis limits.
        zero_at_efermi (bool):
            Whether to show the plot in the absolute scale.
        legend (bool):
            Whether to show the figure legend.
        crop_first_value (bool):
            Whether to crop the fist DOS.
        show_spg (bool):
            Whether to show space group number in the title.
        symprec (float):
            Symprec for determining the equivalent sites.
        angle_tolerance (float):
            Angle tolerance for symmetry analysis in degree.
    """

    v = Vasprun(vasprun, ionic_step_skip=True, parse_eigen=False)
    if v.converged_electronic is False:
        logger.warning("SCF is not attained in the vasp calculation.")

    complete_dos = v.complete_dos

    # check cbm
    if cbm_vbm is None:
        if complete_dos.get_gap() > 0.1:
            cbm_vbm = complete_dos.get_cbm_vbm()

    structure = v.final_structure

    dos = OrderedDict()
    # The CompleteDos behaves as DOS for total dos.
    dos["Total"] = complete_dos

    if specific and specific[0].isdigit():
        if pdos_type is not "none":
            logger.warning(f"pdos_type is changed from {pdos_type} to none")
        pdos_type = "none"

    elif specific and specific[0].isalpha():
        if pdos_type is not "none":
            logger.warning(f"pdos_type is changed from {pdos_type} to element")
        pdos_type = "element"

    sga = None
    grouped_indices = defaultdict(list)

    if pdos_type == "element":
        for indices, s in enumerate(structure):
            grouped_indices[str(s.specie)].append(indices)

    elif pdos_type == "site":
        # equivalent_sites: Equivalent site indices from SpacegroupAnalyzer.
        sga = SpacegroupAnalyzer(structure=structure,
                                 symprec=symprec,
                                 angle_tolerance=angle_tolerance)
        symmetrized_structure = sga.get_symmetrized_structure()
        # equiv_indices = [[0], [1], [2, 3], [4, 5]]
        equiv_index_lists = symmetrized_structure.equivalent_indices

        for l in equiv_index_lists:
            specie = structure[l[0]].specie
            wyckoff = sga.get_symmetry_dataset()["wyckoffs"][l[0]]

            name = f"{specie} {wyckoff}"
            if name in grouped_indices:
                for i in range(2, 10):
                    name = f"{specie} {wyckoff}{i}"
                    if name in grouped_indices:
                        continue
                    else:
                        break
                else:
                    raise InvalidStructureError(
                        "More than 9 same Wyckoff sites are not supported.")
            grouped_indices[name] = l
            logger.info({f"Atoms {grouped_indices[name]} grouped to {name}."})

    elif pdos_type == "none":
        for indices, s in enumerate(structure):
            name = f"{s.specie} site: {indices}"
            grouped_indices[name].append(indices)
    else:
        raise KeyError("The given pdos_type is not supported.")

    if specific:
        tmp = defaultdict(list)
        for key, value in grouped_indices.items():
            if pdos_type == "element" and key in specific:
                tmp[key] = value
            else:
                # type(index) is str
                index = ''.join(c for c in key if c.isdigit())
                if index in specific:
                    tmp[key] = value
        grouped_indices = tmp

    # efermi is set to VBM if exists.
    efermi = cbm_vbm[1] if cbm_vbm else complete_dos.efermi
    complete_dos.efermi = efermi
    energies = complete_dos.energies

    for key, value in grouped_indices.items():
        for indices in value:
            site = structure[indices]
            if orbital:
                for orb, pdos in complete_dos.get_site_spd_dos(site).items():
                    # " " is used for grouping the plots.
                    if pdos_type == "none":
                        name = f"{key} {orb}"
                    else:
                        name = f"{key} #{len(value)} {orb}"
                    density = divide_densities(pdos.densities, len(value))
                    if name in dos:
                        density = add_densities(dos[name].densities, density)
                    dos[name] = Dos(efermi, energies, density)
            else:
                name = f"{key}({len(key)})" if isinstance(key, list) else key
                pdos = complete_dos.get_site_dos(site)
                if name in dos:
                    density = add_densities(dos[name].densities, pdos.densities)
                    dos[name] = Dos(efermi, energies, density)
                else:
                    dos[name] = pdos

    # use complete_dos.efermi for total dos.
    plotter = ViseDosPlotter(zero_at_efermi=zero_at_efermi)
    plotter.add_dos_dict(dos)

    xlim = xlim or [-10, 10]

    if ymaxs:
        if v.incar.get("ISPIN", 1) == 2:
            ylims = [[-y, y] for y in ymaxs]
        else:
            ylims = [[0, y] for y in ymaxs]
    else:
        energies = complete_dos.energies - efermi
        tdos_max = 1.1 * max_density(density=complete_dos.densities,
                                     energies=energies,
                                     xlim=xlim,
                                     crop_first_value=crop_first_value)
        if v.incar.get("ISPIN", 1) == 2:
            ylims = [[-tdos_max, tdos_max]]
        else:
            ylims = [[0, tdos_max]]

        pdos_max = 0.0
        for k, d in dos.items():
            if k == "Total":
                continue
            pdos_max = max(max_density(d.densities, energies, xlim), pdos_max)
        pdos_max *= 1.1

        if v.incar.get("ISPIN", 1) == 2:
            ylims.append([-pdos_max, pdos_max])
        else:
            ylims.append([0, pdos_max])

        logger.info(f"Dos plot, y-range {ylims}")

    comp = latexify(structure.composition.get_reduced_formula_and_factor()[0])
    if show_spg:
        if sga is None:
            sga = SpacegroupAnalyzer(structure, symprec=symprec)
        sg_num_str = str(sga.get_space_group_number())
        sg_symbol = latexify_spacegroup(sga.get_space_group_symbol())
        sg = f" {sg_symbol} ({sg_num_str})"
        logger.info(f"Space group: {sg}")
        title = f"{comp}, SG: {sg}"
    else:
        title = comp

    return plotter.get_plot(xlim=xlim,
                            ylims=ylims,
                            cbm_vbm=cbm_vbm,
                            legend=legend,
                            crop_first_value=crop_first_value,
                            title=title)


def divide_densities(density: dict,
                     denominator: float) -> Dict[Spin, np.ndarray]:
    """Method to divide density.

    Args:
        density:
            Density of states as a function of Spin.
        denominator:
            Denominator.
    Returns:
        Density of states with the format of {spin: np.array(density)}.
    """

    return {spin: np.array(val) / denominator for spin, val in density.items()}


def max_density(density: dict,
                energies: list,
                xlim: list,
                crop_first_value: bool = True) -> float:
    """Evaluate max value of the density of states in the given energy range.

    Args:
        density (dict):
            Note that the first value may contain a huge value when the
            lower limit of the calculation of density of states is larger than
            that of occupied states. Therefore, we need to crop the first value
            by default.
            {Spin.up: [...], Spin.down: [...] }
        energies (list):
            Energy mesh values.
        xlim (list):
             x-range with [x-min, x-max]
        crop_first_value (bool):
            Whether to crop the first value or not.

    Return (float):
        Max value in the density within the given x-range.
    """
    values = []
    for density_in_each_spin in density.values():
        for i, (d, e) in enumerate(zip(density_in_each_spin, energies)):
            if crop_first_value and i == 0:
                continue
            if xlim[0] < e < xlim[1]:
                values.append(d)

    if not values:
        raise ValueError("DOS is empty at the given energy {0[0]} - {0[1]} "
                         "range.".format(xlim))

    return max(values)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

from typing import List, Union

import numpy as np

import scipy.interpolate as scint

from matplotlib import pyplot

from pymatgen.electronic_structure.bandstructure import BandStructureSymmLine
from pymatgen.electronic_structure.bandstructure import (
    get_reconstructed_band_structure)
from pymatgen.electronic_structure.core import Spin
from pymatgen.electronic_structure.plotter import BSPlotter
from pymatgen.io.vasp import Kpoints, Vasprun
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from pymatgen.util.plotting import pretty_plot
from pymatgen.util.string import latexify_spacegroup, latexify

from vise.config import SYMMETRY_TOLERANCE, ANGLE_TOL
from vise.util.logger import get_logger


logger = get_logger(__name__)


class PrettyBSPlotter:

    def __init__(self,
                 band: BandStructureSymmLine,
                 band2: BandStructureSymmLine = None,
                 absolute: bool = False,
                 y_range: List[float] = None,
                 legend: bool = False,
                 symprec: float = SYMMETRY_TOLERANCE,
                 angle_tolerance: float = ANGLE_TOL) -> None:
        """

        Note:

        Args:
            band (BandStructureSymmLine):
                A BandStructureSymmLine object.
            band2 (BandStructureSymmLine):
                Another BandStructureSymmLine object to be compared.
            absolute (bool):
                Whether to show the band structure in the absolute energy scale.
            y_range (List[float, float]):
                The min and max of energy range.
            legend (bool):
                Whether to show te figure legend.
            symprec (float):
                Symprec used for determining the space group.
            angle_tolerance (float):
                Angle tolerance used for determining the space group.
        """

        bs_plotter = ModBSPlotter(band)
        comp = band.structure.composition.get_reduced_formula_and_factor()[0]
        sga = SpacegroupAnalyzer(structure=band.structure,
                                 symprec=symprec,
                                 angle_tolerance=angle_tolerance)

        comp = latexify(comp)
        sg_symbol = latexify_spacegroup(sga.get_space_group_symbol())
        sg_num = sga.get_space_group_number()

        kwargs = {"ylim": y_range,
                  "legend": legend,
                  "zero_to_efermi": not absolute,
                  "title": f"{comp}, SG: {sg_symbol} ({sg_num})"}

        if not band2:
            self.plotter = bs_plotter.get_plot(**kwargs)
        else:
            bs_plotter2 = ModBSPlotter(band2)
            self.plotter = bs_plotter2.plot_compare(bs_plotter, **kwargs)

    @classmethod
    def from_vasp_files(cls,
                        kpoints_filenames: Union[List[str], str],
                        vasprun_filenames: Union[List[str], str],
                        vasprun2_filenames: Union[List[str], str] = None,
                        **kwargs) -> "PrettyBSPlotter":

        band = cls.make_sym_line(kpoints_filenames, vasprun_filenames)
        if vasprun2_filenames:
            band2 = cls.make_sym_line(kpoints_filenames, vasprun2_filenames)
        else:
            band2 = None

        return cls(band=band, band2=band2, **kwargs)

    @staticmethod
    def make_sym_line(kpoints_filenames: Union[List[str], str],
                      vasprun_filenames: Union[List[str], str]
                      ) -> "VaspBandStructureSymmLine":
        """Create VaspBandStructureSymmLine from vasp files.


        """

        if isinstance(kpoints_filenames, list):
            band = [VaspBandStructureSymmLine(k, v)
                    for k, v in zip(kpoints_filenames, vasprun_filenames)]
            return get_reconstructed_band_structure(band)
        else:
            return VaspBandStructureSymmLine(kpoints_filenames,
                                             vasprun_filenames)

    def show(self, filename: str = None, format_type: str = "pdf") -> None:
        if filename:
            self.plotter.savefig(filename, format=format_type)
        else:
            self.plotter.show()


class ModBSPlotter(BSPlotter):

    def get_plot(self,
                 ylim: List[float] = None,
                 smooth: bool = False,
                 vbm_cbm_marker: bool = True,
                 smooth_tol: float = None,
                 legend: bool = False,
                 zero_to_efermi: bool = True,
                 title: str = None) -> pyplot:
        """Get a matplotlib object for the band structure plot.

        Original function is copied from PYMATGEN.2018.5.22.

        Args:
            ylim (list):
                Specify the y-axis (energy) limits; by default None let the code
                choose. It is vbm-4 and cbm+4 if insulator efermi-10 and
                efermi+10 if metal.
            zero_to_efermi (bool):
                Automatically subtract off the Fermi energy from the eigenvalues
                and plot (E-Ef).
            smooth (bool):
                Whether to interpolates the bands by a spline cubic
            vbm_cbm_marker (bool):
                Whether to show the VBM and CBM markers.
            smooth_tol (float):
                Tolerance for fitting spline to band data. Default is None such
                that no tolerance will be used.
            legend (bool):
                If the figure legend is shown or not.
            title (str):
                The title of the figure.

        Returns:
            Matplotlib pyplot
        """
        plt = pretty_plot(12, 8)

        band_line_width = 1
        spin = [Spin.up, Spin.down] if self._bs.is_spin_polarized else [Spin.up]
        line_style = {Spin.up: 'b-', Spin.down: 'r--'}

        # Ref:
        # http://pymatgen.org/modules/pymatgen/electronic_structure/plotter.html
        data = self.bs_plot_data(zero_to_efermi)
        if not smooth:
            for d in range(len(data['distances'])):
                for i in range(self._nb_bands):
                    for s in spin:
                        data_x = data['distances'][d]
                        data_y = [data['energy'][d][str(s)][i][j]
                                  for j in range(len(data['distances'][d]))]
                        plt.plot(data_x, data_y, line_style[s],
                                 linewidth=band_line_width)
        else:
            # Interpolation failure can be caused by trying to fit an entire
            # band with one spline rather than fitting with piecewise splines
            # (splines are ill-suited to fit discontinuities).
            #
            # The number of splines used to fit a band is determined by the
            # number of branches (high symmetry lines) defined in the
            # BandStructureSymmLine object (see BandStructureSymmLine._branches)

            warning = "WARNING! Distance / branch {d}, band {i} cannot be " \
                      "interpolated.\n See full warning in source.\n " \
                      "If this is not a mistake, try increasing smooth_tol.\n"\
                      "Current smooth_tol is {s}."

            for d in range(len(data['distances'])):
                # self._nb_bands means the number of bands.
                for i in range(self._nb_bands):
                    for s in spin:
                        # construct splines for each branch, e.g., Gamma-K
                        tck = scint.splrep(
                            data['distances'][d],
                            [data['energy'][d][str(s)][i][j]
                             for j in range(len(data['distances'][d]))],
                            s=smooth_tol)
                        step = (data['distances'][d][-1]
                                - data['distances'][d][0]) / 1000
                        xs = [x * step + data['distances'][d][0]
                              for x in range(1000)]
                        ys = [scint.splev(
                            x * step + data['distances'][d][0], tck, der=0)
                            for x in range(1000)]

                        for y in ys:
                            if np.isnan(y):
                                print(warning.format(d=str(d), i=str(i),
                                                     s=str(smooth_tol)))
                            break

                        plt.plot(xs, ys, line_style[s],
                                 linewidth=band_line_width)

        self._maketicks(plt)

        # Main X and Y Labels
        plt.xlabel(r'$\mathrm{Wave\ vector}$', fontsize=30)
        plt.ylabel(r'$\mathrm{Energy\ (eV)}$', fontsize=30)

        # Draw Fermi energy, only if not zero and metal
        if not zero_to_efermi and self._bs.is_metal():
            ef = self._bs.efermi
            plt.axhline(ef, linewidth=0.5, color='k')

        # X max range = last distance point
        x_max = data['distances'][-1][-1]
        plt.xlim(0, x_max)

        if ylim:
            plt.ylim(ylim)
        else:
            if self._bs.is_metal():
                fl = self._bs.efermi
                e_range = [fl - 10, fl + 10] if zero_to_efermi else [-10, 10]
                plt.ylim(e_range)
            else:
                plt.ylim(data['vbm'][0][1] - 4, data['cbm'][0][1] + 4)

        if not self._bs.is_metal() and vbm_cbm_marker:
            for cbm in data['cbm']:
                plt.scatter(cbm[0], cbm[1], color='r', marker='o', s=100)
            for vbm in data['vbm']:
                plt.scatter(vbm[0], vbm[1], color='g', marker='o', s=100)

        if legend:
            import matplotlib.lines as mlines
            if self._bs.is_spin_polarized:
                handles = [mlines.Line2D([], [], linewidth=2, color='b',
                                         label='bs 1 up'),
                           mlines.Line2D([], [], linewidth=2, color='r',
                                         label='bs 1 down', linestyle="--")]
            else:
                handles = [mlines.Line2D([], [], linewidth=2, color='b',
                                         label='bs 1')]
            plt.legend(handles=handles)

        plt.tight_layout()

        # plot the band gap with an arrow.
        if self._bs.is_metal() is False:
            band_gap = self._bs.get_band_gap()["energy"]
            x_position = (plt.xlim()[0] + plt.xlim()[1]) / 2
            vbm = self.bs_plot_data(zero_to_efermi=zero_to_efermi)["vbm"][0][1]
            plt = self.add_band_gap(plt, vbm, band_gap, x_position)

        if title:
            plt.title(title, size=20)

        return plt

    @staticmethod
    def add_band_gap(plt: pyplot,
                     vbm: float,
                     band_gap: float,
                     annotate_x: float,
                     line_color: str = "purple") -> pyplot:
        plt.hlines([vbm, vbm + band_gap], plt.xlim()[0], plt.xlim()[1],
                   line_color, linestyles='dashed')

        plt.annotate('', xy=(annotate_x, vbm),
                     xytext=(annotate_x, vbm + band_gap),
                     fontsize=11, color='black',
                     arrowprops=dict(edgecolor='black', arrowstyle='<|-|>',
                                     shrinkA=0, shrinkB=0))
        plt.annotate(f"{str(round(band_gap, 3))} eV",
                     (annotate_x * 1.03, vbm + band_gap * 0.4), fontsize=20)

        return plt

    def plot_compare(self,
                     another_plotter: "ModBSPlotter",
                     ylim: List[float] = None,
                     legend: bool = True,
                     zero_to_efermi: bool = True,
                     title: str = None) -> pyplot:
        """Generate pyplot for comparing two band structures.

        Note: x- and y-ranges are determined based on the original data.

        Args:
            another_plotter (ModBSPlotter):
                Another BSPlotter class object.
            ylim (List[float]):
                Y-axis limit [y min, y max].
            legend (bool):
                If the figure legend is shown or not.
            zero_to_efermi (bool):
                If the energy is aligned to the Fermi level or not.
            title (str):
                Title of the plot.

        Returns:
            Matplotlib pyplot.
        """

        plt = self.get_plot(ylim=ylim, zero_to_efermi=zero_to_efermi)
        data_orig = self.bs_plot_data(zero_to_efermi=zero_to_efermi)
        data = another_plotter.bs_plot_data(zero_to_efermi=zero_to_efermi)

        spin = [Spin.up, Spin.down] if self._bs.is_spin_polarized else [Spin.up]

        second_line_style = {Spin.up: 'c-', Spin.down: 'm--'}

        # plot second band
        second_band_line_width = 2
        for i in range(another_plotter._nb_bands):
            for d in range(len(data_orig['distances'])):
                for s in spin:
                    data_x = data_orig['distances'][d]
                    data_y = [e[str(s)][i] for e in data['energy']][d]
                    plt.plot(data_x, data_y, second_line_style[s],
                             linewidth=second_band_line_width)

        # add second band gap and its arrow.
        if another_plotter._bs.is_metal() is False:
            band_gap = another_plotter._bs.get_band_gap()["energy"]
            x = (plt.xlim()[0] * 0.35 + plt.xlim()[1] * 0.65)
            vbm = another_plotter.bs_plot_data(
                zero_to_efermi=zero_to_efermi)["vbm"][0][1]
            plt = self.add_band_gap(plt, vbm, band_gap, x)

        import matplotlib.lines as mlines
        if legend:
            if self._bs.is_spin_polarized:
                handles = [mlines.Line2D([], [], linewidth=2, color='b',
                                         label='bs 1 up'),
                           mlines.Line2D([], [], linewidth=2, color='r',
                                         label='bs 1 down', linestyle="--")]
            else:
                handles = [mlines.Line2D([], [], linewidth=2, color='b',
                                         label='bs 1')]
            if another_plotter._bs.is_spin_polarized:
                handles.extend([mlines.Line2D([], [], linewidth=2, color='c',
                                              label='bs 2 up'),
                                mlines.Line2D([], [], linewidth=2, color='m',
                                              linestyle="--",
                                              label='bs 2 down')])
            else:
                handles.extend([mlines.Line2D([], [], linewidth=2, color='c',
                                              label='bs 2')])
            plt.legend(handles=handles)

        if title:
            plt.title(title, size=20)

        return plt


class VaspBandStructureSymmLine(BandStructureSymmLine):

    def __init__(self,
                 kpoints_filename: str,
                 vasprun_filename: str,
                 is_projection=False):
        """Create BandStructureSymmLine from KPOINTS and vasprun.xml.

        Notes: KPOINTS file includes finite weight part and zero weight part;
               the former is used for constructing the Hamiltonian while the
               latter for band structure symmetry line.

        Args:
            kpoints_filename (str):
                KPOINTS file name parsed.
            vasprun_filename (str):
                vasprun.xml file name parsed.
            is_projection (bool):
                Whether to calculate the projection values.
        """
        kpoints = Kpoints.from_file(kpoints_filename)

        # BSVasprun cannot be used when the electronic convergence is checked.
        vasprun = Vasprun(filename=vasprun_filename,
                          parse_projected_eigen=is_projection)

        if vasprun.converged_electronic is False:
            logger.critical("SCF is not attained!!!!!")

        eigenvalues = vasprun.eigenvalues
        reciprocal_lattice = vasprun.final_structure.lattice.reciprocal_lattice
        efermi = vasprun.efermi

        # K-points with weight for Hamiltonian are omitted from the plot.
        first_index_wo_weight = sum(w > 0 for w in kpoints.kpts_weights)
        kpts_wo_weight = kpoints.kpts[first_index_wo_weight:]

        eigenvalues_wo_weight = {}
        for s in eigenvalues:
            # transpose is essential
            # When parsing vasprun.xml using BSVasprun,
            # eigenvalues[kpt-index][band-index] = [energy, occupation]
            # For BSPlotter
            # eigenvalues[band-index][kpt-index] = energy
            eigenvalues_wo_weight[Spin(s)] = \
                eigenvalues[Spin(s)][first_index_wo_weight:, :, 0].transpose()

        # Store label except for "None".
        labels_dict = {}
        for i, label in enumerate(kpoints.labels):
            if label != "None" and label is not None:
                if label == "GAMMA":
                    labels_dict["\u0393"] = kpoints.kpts[i]
                elif label == "SIGMA_0":
                    labels_dict["\u03A3" + "_0"] = kpoints.kpts[i]
                elif label == "DELTA_0":
                    labels_dict["\u0394" + "_0"] = kpoints.kpts[i]
                else:
                    labels_dict[label] = kpoints.kpts[i]

        structure = vasprun.final_structure

        # TODO: Write unittest for the projection.
        # TODO: Use projections for the plot.
        projections = None
        if is_projection:
            v = vasprun.projected_eigenvalues
            projections = {}
            for s in eigenvalues:
                # swapaxes is essential
                # When parsing vasprun.xml using BSVasprun,
                #  Vasprun.projected_eigenvalues[spin][kpoint index][band index]
                #                               [atom index][orbital_index]
                # For BSPlotter
                # projections: dict of orbital projections as {spin: ndarray}.
                # The indices of the ndarray are [band_index, kpoint_index,
                #                                 orbital_index, ion_index].
                projections[Spin(s)] = v[Spin(s)].swapaxes(0, 1).swapaxes(2, 3)

        super().__init__(kpoints=kpts_wo_weight,
                         eigenvals=eigenvalues_wo_weight,
                         lattice=reciprocal_lattice,
                         efermi=efermi,
                         labels_dict=labels_dict,
                         structure=structure,
                         projections=projections)



# -*- coding: utf-8 -*-

import os
from pathlib import Path
import shutil

from pymatgen.core.composition import Composition
from pymatgen.core.sites import Element

from vise.chempotdiag.free_energy_entries import (
    FreeEnergyEntry, FreeEnergyEntrySet, ConstrainedFreeEnergyEntrySet)
from vise.util.testing import ViseTest


DISABLE_DISPLAY_DIAGRAM = False
parent_dir = Path(__file__).parent


class TestFreeEnergyEntry(ViseTest):
    def setUp(self) -> None:
        self.mgo = FreeEnergyEntry("MgO", total_energy=-10)
        self.o2 = FreeEnergyEntry("O2",
                                  total_energy=-2,
                                  zero_point_vib=-10,
                                  free_e_shift=-100,
                                  data={"id": 10},
                                  attribute="new")

    def test_dict(self):
        expected = self.o2.as_dict()
        actual = FreeEnergyEntry.from_dict(expected).as_dict()
        self.assertEqual(expected, actual)

    def test_msonable(self):
        self.assertMSONable(self.mgo)
        self.assertMSONable(self.o2)

    def test_print(self):
        expected = "PDEntry : O2 with composition O2. Energy: -112.000. Zero point vib energy: -10.000. Free energy contribution: -100.000. Data: {'id': 10}"
        self.assertEqual(expected, str(self.o2))


class TestFreeEnergyEntrySet(ViseTest):
    def setUp(self) -> None:
        mgo = FreeEnergyEntry("MgO", total_energy=-10)
        self.mg = FreeEnergyEntry("Mg", total_energy=-10)
        self.o2 = FreeEnergyEntry("O2",
                                  total_energy=-2,
                                  zero_point_vib=-10,
                                  free_e_shift=-100,
                                  data={"id": 10},
                                  attribute="new")

        self.entry_set = FreeEnergyEntrySet([mgo, self.o2])

    def test_msonable(self):
        self.assertMSONable(self.entry_set)

    def test_add_discard(self):
        es = self.entry_set
        self.assertEqual(2, len(es))
        es.add(self.mg)
        self.assertEqual(3, len(es))
        es.discard(self.o2)
        self.assertEqual(2, len(es))

    def test_csv(self):
        self.entry_set.to_csv("entries.csv")
        actual = open("entries.csv").read()
        expected = """Name,Mg,O,Energy
MgO,1.0,1.0,-10
O2,0,2.0,-112
"""
        self.assertEqual(expected, actual)
        os.remove("entries.csv")

    def test_from_vasp_files(self):
        paths = [parent_dir / "vasp_Mg",
                 parent_dir / "mol_O2",
                 parent_dir / "vasp_MgO"]
        entry_set = FreeEnergyEntrySet.from_vasp_files(paths,
                                                       parse_gas=True,
                                                       temperature=200)
        expected = {"Mg", "MgO", "O2"}
        actual = {str(e.name) for e in entry_set}

        self.assertEqual(expected, actual)
        expected = {-5.18827492, -12.51242284, -9.20808429}
        actual = {e.energy for e in entry_set}
        self.assertEqual(expected, actual)
        print(entry_set)

    def test_from_mp(self):
        entry_set = FreeEnergyEntrySet.from_mp(["Mg", "O"])
        expected = {"Mg", "MgO", "O2"}
        actual = {str(e.name) for e in entry_set}
        self.assertEqual(expected, actual)

        expected = {-39.42964153, -14.43901464, -11.96804153}
        actual = {e.energy for e in entry_set}
        self.assertEqual(expected, actual)
        print(entry_set)

    def test_dict(self):
        expected = self.o2.as_dict()
        actual = FreeEnergyEntry.from_dict(expected).as_dict()
        self.assertEqual(expected, actual)

    def test_msonable(self):
        self.assertMSONable(self.entry_set)

    def test_print(self):
        print(self.entry_set)


class TestConstrainedFreeEnergyEntrySet(ViseTest):
    def setUp(self) -> None:

        mg = FreeEnergyEntry("Mg", total_energy=-1)
        ca = FreeEnergyEntry("Ca", total_energy=-1)
        mgo = FreeEnergyEntry("MgO", total_energy=-12)
        cao = FreeEnergyEntry("CaO", total_energy=-12)
        o2 = FreeEnergyEntry("O2", total_energy=-2)
        mgcao = FreeEnergyEntry("MgCaO", total_energy=-33)
        self.entry_set = FreeEnergyEntrySet([mg, ca, mgo, cao, o2, mgcao])

    def test_constraint(self):
        entry_set = ConstrainedFreeEnergyEntrySet.\
            from_entry_set(self.entry_set, {"Ca": -2})
        expected = {Element.Mg, Element.O}
        actual = set()
        for e in entry_set.entries:
            actual.update(e.composition.elements)
        self.assertEqual(expected, actual)

        expected = FreeEnergyEntry(composition=Composition({"Mg": 1, "O": 1}),
                                   total_energy=-33-(-2),
                                   name="CaMgO")
        actual = [e for e in entry_set if e.name == "CaMgO"][0]
        self.assertEqual(str(expected), str(actual))





# -*- coding: utf-8 -*-

from math import ceil
from pathlib import Path
from typing import Optional

from monty.serialization import loadfn

from pymatgen import Structure, Composition, Element
from pymatgen.io.vasp import Potcar

from vise.config import BAND_GAP_CRITERION
from vise.input_set.settings_util import (
    load_default_incar_settings, check_keys, nelect, nbands, calc_npar_kpar)
from vise.input_set.task import (
    LATTICE_RELAX_TASK, PLOT_TASK, SPECTRA_TASK, Task)
from vise.input_set.xc import Xc, LDA_OR_GGA, HYBRID_FUNCTIONAL
from vise.util.logger import get_logger

logger = get_logger(__name__)

SET_DIR = Path(__file__).parent / "datasets"

TASK_REQUIRED_FLAGS = {"LREAL", "ISPIN", "ISIF", "EDIFF", "IBRION", "ISMEAR",
                       "PREC"}
TASK_OPTIONAL_FLAGS = {"NSW", "EDIFFG", "POTIM", "ADDGRID", "KPAR", "NPAR",
                       "ENCUT", "NBANDS", "LEPSILON", "LCALCEPS", "LOPTICS",
                       "CSHIFT", "EMIN", "EMAX", "NEDOS"}
TASK_FLAGS = TASK_REQUIRED_FLAGS | TASK_OPTIONAL_FLAGS

XC_REQUIRED_FLAGS = {"ALGO", "LWAVE"}
XC_OPTIONAL_FLAGS = {"GGA", "METAGGA", "LDAU", "LDAUTYPE", "LDAUPRINT",
                     "LDAUU", "LDAUL", "LMAXMIX", "NKRED", "LHFCALC",
                     "PRECFOCK", "TIME", "HFSCREEN", "AEXX", "NKRED"}
XC_FLAGS = XC_REQUIRED_FLAGS | XC_OPTIONAL_FLAGS

XC_TASK_REQUIRED_FLAGS = set()
XC_TASK_OPTIONAL_FLAGS = set()
XC_TASK_FLAGS = XC_TASK_REQUIRED_FLAGS | XC_TASK_OPTIONAL_FLAGS

COMMON_REQUIRED_FLAGS = {"NELM", "LASPH", "LORBIT", "LCHARG", "SIGMA"}
COMMON_OPTIONAL_FLAGS = {"NELECT"}
COMMON_FLAGS = COMMON_REQUIRED_FLAGS | COMMON_OPTIONAL_FLAGS

ALL_FLAGS = set(sum(loadfn(SET_DIR / "incar_flags.yaml").values(), []))

OTHER_FLAGS = ALL_FLAGS - (TASK_FLAGS | XC_FLAGS | XC_TASK_FLAGS | COMMON_FLAGS)


# TODO: Implement spin-orbit calc.

class TaskIncarSettings:

    def __init__(self,
                 settings: dict):

        check_keys(settings, TASK_REQUIRED_FLAGS, TASK_OPTIONAL_FLAGS)
        self.settings = settings

    @classmethod
    def from_options(cls,
                     task: Task,
                     structure: Structure,
                     potcar: Potcar,
                     num_kpoints: int,
                     max_enmax: float,
                     is_magnetization: bool,
                     vbm_cbm: Optional[list],
                     npar_kpar: bool,
                     num_nodes: Optional[int],
                     encut: Optional[float],
                     structure_opt_encut_factor: float,
                     dos_step_size: float) -> "TaskIncarSettings":
        """Construct incar settings related to task with some options.

        Args: See ViseInputSet docstrings.

        Returns:
            TaskIncarSettings instance object.
        """
        band_gap = vbm_cbm[1] - vbm_cbm[0] if vbm_cbm else None
        if band_gap:
            logger.info(f"Band gap: {round(band_gap, 3)} eV.")
            is_band_gap = band_gap > BAND_GAP_CRITERION
        else:
            is_band_gap = None

        settings = \
            load_default_incar_settings(yaml_filename="task_incar_set.yaml",
                                        required_flags=TASK_REQUIRED_FLAGS,
                                        optional_flags=TASK_OPTIONAL_FLAGS,
                                        key_name=str(task))

        settings["LREAL"] = "A" if task == Task.defect else False

        # -5 is acceptable for num_kpt >= 4 and is_band_gap = True
        if is_band_gap is not True or num_kpoints < 4:
            settings["ISMEAR"] = 0

        if task == Task.defect or is_magnetization:
            settings["ISPIN"] = 2
            if hasattr(structure, "site_properties"):
                if "magmom" in structure.site_properties:
                    settings["MAGMOM"] = structure.site_properties["magmom"]

        if npar_kpar:
            kpar, npar = calc_npar_kpar(num_kpoints, num_nodes)
            settings["KPAR"] = kpar
            # now switch off NPAR
#            settings["NPAR"] = npar

        if not encut:
            if task in LATTICE_RELAX_TASK:
                encut = round(max_enmax * structure_opt_encut_factor, 3)
            else:
                encut = max_enmax
        settings["ENCUT"] = encut

        if task in PLOT_TASK:
            settings["NBANDS"] = nbands(structure.composition, potcar)

        if task in SPECTRA_TASK:
            if vbm_cbm:
                emin = ceil(vbm_cbm[0]) - 15 - dos_step_size
                emax = ceil(vbm_cbm[1]) + 15
            else:
                emin = -20 - dos_step_size
                emax = 20
            nedos = int(round((emax - emin) / dos_step_size, 0)) + 1
            settings.update({"EMIN": emin, "EMAX": emax, "NEDOS": nedos})

        if task == Task.dielectric_dfpt:
            settings["LEPSILON"] = True
        elif task == Task.dielectric_finite_field:
            settings["LCALCEPS"] = True
            settings["IBRION"] = 6
            settings["POTIM"] = 0.015
        elif task == Task.dielectric_function:
            settings["LOPTICS"] = True
            settings["CSHIFT"] = 0.01

        return cls(settings=settings)


class XcIncarSettings:

    def __init__(self, settings: dict):
        check_keys(settings, XC_REQUIRED_FLAGS, XC_OPTIONAL_FLAGS)
        self.settings = settings

    @classmethod
    def from_options(cls,
                     xc: Xc,
                     symbol_list: list,
                     factor: int,
                     aexx: Optional[float] = 0.25,
                     hubbard_u: Optional[bool] = None,
                     ldauu: Optional[dict] = None,
                     ldaul: Optional[dict] = None,
                     ldaul_set_name: Optional[str] = "default"
                     ) -> "XcIncarSettings":
        """Construct incar settings related to xc with some options.

        Args: See ViseInputSet docstrings

        Returns:
            XcIncarSettings instance object.
        """
        settings = \
            load_default_incar_settings(yaml_filename="xc_incar_set.yaml",
                                        required_flags=XC_REQUIRED_FLAGS,
                                        optional_flags=XC_OPTIONAL_FLAGS,
                                        key_name=str(xc))

        if ldauu:
            hubbard_u = True
        # By default Hubbard U is set for LDA or GGA.
        hubbard_u = xc in LDA_OR_GGA if hubbard_u is None else hubbard_u
        ldauu = ldauu or {}
        ldaul = ldaul or {}

        if xc == Xc.pbesol:
            settings["GGA"] = "PS"
        elif xc == Xc.scan:
            settings["METAGGA"] = "SCAN"

        if hubbard_u:
            u_set = loadfn(SET_DIR / "u_parameter_set.yaml")
            ldauu_set = u_set["LDAUU"][ldaul_set_name]
            ldauu_set.update(ldauu)
            ldauu = [ldauu_set.get(el, 0) for el in symbol_list]

            if sum(ldauu) > 0:
                settings["LDAUU"] = ldauu
                settings.update({"LDAU": True, "LDAUTYPE": 2, "LDAUPRINT": 1})
                ldaul_set = u_set["LDAUL"][ldaul_set_name]
                ldaul_set.update(ldaul)
                settings["LDAUL"] = \
                    [ldaul_set.get(el, -1) for el in symbol_list]
                settings["LMAXMIX"] = \
                    6 if any([Element(el).Z > 56 for el in symbol_list]) else 4

        if xc in HYBRID_FUNCTIONAL:
            settings["AEXX"] = aexx
            if factor > 1:
                settings["NKRED"] = factor

        return cls(settings=settings)


class XcTaskIncarSettings:

    def __init__(self, settings: dict):
        check_keys(settings, XC_TASK_REQUIRED_FLAGS, XC_TASK_OPTIONAL_FLAGS)
        self.settings = settings

    @classmethod
    def from_options(cls) -> "XcTaskIncarSettings":
        """Construct incar settings related to task and xc with options.

        Note: When the GW related task and xc are implemented, this class will
              be used.

        Args: See ViseInputSet docstrings

        Returns:
            XcTaskIncarSettings instance object.
        """
        settings = {}
        return cls(settings)


class CommonIncarSettings:
    def __init__(self, settings: dict):
        check_keys(settings, COMMON_REQUIRED_FLAGS, COMMON_OPTIONAL_FLAGS)
        self.settings = settings

    @classmethod
    def from_options(cls,
                     potcar: Potcar,
                     composition: Composition,
                     charge: Optional[int] = None) -> "CommonIncarSettings":
        """Construct incar settings related to task and xc with options.

        Args: See ViseInputSet docstrings

        Returns:
            CommonIncarSettings instance object.
        """
        settings = {"NELM": 100,
                    "LASPH": True,
                    "LORBIT": 12,
                    "LCHARG": False,
                    "SIGMA": 0.1}
        # Structure charge is ignored.
        if charge:
            settings["NELECT"] = nelect(composition, potcar, charge)

        return cls(settings=settings)



# -*- coding: utf-8 -*-

from enum import Enum, unique
from math import ceil, modf, pow
from typing import List

import numpy as np
from pymatgen import Structure
from pymatgen.core.periodic_table import Element
from pymatgen.io.vasp.inputs import Kpoints
from pymatgen.symmetry.analyzer import SpacegroupAnalyzer
from vise.config import BAND_REF_DIST, KPT_DENSITY
from vise.config import SYMMETRY_TOLERANCE, ANGLE_TOL
from vise.input_set.datasets.kpt_centering import kpt_centering
from vise.util.logger import get_logger
from vise.util.structure_handler import (
    structure_to_seekpath, find_spglib_primitive, get_symmetry_dataset)

logger = get_logger(__name__)

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


@unique
class KpointsMode(Enum):
    """K-point generation type

       Supporting modes are:
       "band":
           Kpoints with the band path will be returned based on the
           seekpath program. The space group is analyzed and primitive
           unitcell that must be used for the band structure calculation is
           returned as well.
       "primitive_uniform":
           Kpoints with uniform k-point sampling. The k-point sampling mesh
           and centering are determined based on the standardized primitive
           unitcell. Structure is also changed if not primitive.
       "manual_set":
           Kpoints with uniform k-point sampling. The k-point sampling mesh
           and centering are determined based on the given lattice. Note
           that only when the angles are 90 degrees, the centering is
           shifted along the perpendicular direction.
           This mode is useful when calculating the supercells.
    """

    band = "band"
    primitive_uniform = "primitive_uniform"
    manual_set = "manual_set"

    def __repr__(self):
        return self.value

    # Don't remove __str__. This is a must.
    def __str__(self):
        return self.value

    @classmethod
    def from_string(cls, s: str):
        for m in cls:
            if m.value == s or m.name == s:
                return m
        raise AttributeError(f"k-point mode {str(s)} is not proper.\n",
                             f"Supported modes:\n {cls.name_list()}")

    @classmethod
    def name_list(cls):
        return ', '.join([e.name for e in cls])


class MakeKpoints:
    """Make Kpoint based on default settings depending on the task.

        # The structures of aP (SG:1, 2), mC (5, 8, 9, 12, 15) and
        # oA (38, 39, 40, 41) are different between spglib and seekpath.
        # see Y. Hinuma et al. Comput. Mater. Sci. 128 (2017) 140–184
        # -- spglib mC
        #  6.048759 -3.479491 0.000000
        #  6.048759  3.479491 0.000000
        # -4.030758  0.000000 6.044512
        # -- seekpath mC
        #  6.048759  3.479491  0.000000
        # -6.048759  3.479491  0.000000
        # -4.030758  0.000000  6.044512
        # -- spglib oA
        #  6.373362  0.000000  0.000000
        #  0.000000  3.200419  5.726024
        #  0.000000 -3.200419  5.726024
        # -- seekpath oA
        #  0.000000  3.200419 -5.726024
        #  0.000000  3.200419  5.726024
        #  6.373362  0.000000  0.000000
    """

    def __init__(self,
                 mode: str,
                 structure: Structure,
                 kpt_density: float = KPT_DENSITY,
                 only_even: bool = False,
                 manual_kpts: list = None,
                 ref_distance: float = BAND_REF_DIST,
                 kpt_shift: list = None,
                 factor: int = 1,
                 symprec: float = SYMMETRY_TOLERANCE,
                 angle_tolerance: float = ANGLE_TOL,
                 is_magnetization: bool = False):
        """Kpoint object based on default settings depending on the task.

        Note that it does not check if primitive cell is standardized or not.

        Args:
            mode (str):
                KpointsMode in string.
            structure (Structure/IStructure):
                An input Structure object.
            kpt_density (float):
                Density of k-point mesh along each direction.
            only_even (bool):
                Numbered k points are ceiled to be even numbers.
            manual_kpts (3x1 list):
                Manual set of the numbers of k-points.
            ref_distance (float):
                Distance for the k-point mesh.
            kpt_shift (1x3 list):
                K-point shift in the definition of the vasp setting.
            factor (int):
                Multiplier factor. This is useful for the use of NKRED = factor
                for hybrid xc calculations.
            symprec (float):
                Precision in Angstrom used for the symmetry search.
            angle_tolerance (float):
                Angle tolerance used for symmetry analyzer.
            is_magnetization (bool):
                Whether the magnetization is considered or not.
                This modifies the band structure path of systems w/o inversion.
        """
        self.mode = KpointsMode.from_string(mode)
        self.initial_structure = structure
        self.kpt_density = kpt_density
        self.only_even = only_even
        self.manual_kpts = manual_kpts
        self.ref_distance = ref_distance
        self.kpt_shift = kpt_shift
        self.factor = factor
        self.symprec = symprec
        self.angle_tolerance = angle_tolerance
        self.is_magnetization = is_magnetization

        self.comment = None
        self.kpt_mesh = None
        self.kpoints = None
        self.num_kpts = None
        self.corresponding_structure = None
        self.sg = None
        self.sg_symbol = None

        if self.mode == KpointsMode.band:
            self.seekpath_info = \
                structure_to_seekpath(structure=structure,
                                      ref_distance=ref_distance,
                                      time_reversal=(not self.is_magnetization),
                                      symprec=symprec,
                                      angle_tolerance=angle_tolerance)
        else:
            self.seekpath_info = None

    def make_kpoints(self):
        self._set_structure()
        self._set_kmesh()
        if self.kpt_shift is None:
            self._set_centering()

        self.kpoints = Kpoints(comment=self.comment,
                               kpts=(self.kpt_mesh,),
                               kpts_shift=self.kpt_shift)
        if self.mode in (KpointsMode.primitive_uniform, KpointsMode.manual_set):
            self.num_kpts = \
                num_irreducible_kpoints(kpoints=self.kpoints,
                                        structure=self.corresponding_structure,
                                        symprec=self.symprec,
                                        angle_tolerance=self.angle_tolerance)
        else:
            irreducible_kpts = \
                irreducible_kpoints(structure=self.corresponding_structure,
                                    kpoints=self.kpoints,
                                    symprec=self.symprec,
                                    angle_tolerance=self.angle_tolerance)
            kpts = [k[0] for k in irreducible_kpts]
            weights = [k[1] for k in irreducible_kpts]
            labels = [None] * len(irreducible_kpts)
            self.kpoints = Kpoints(comment=self.comment,
                                   style=Kpoints.supported_modes.Reciprocal,
                                   num_kpts=len(kpts),
                                   kpts=kpts,
                                   kpts_weights=weights,
                                   labels=labels)
            self._add_band_kpts()
            self.num_kpts = len(self.kpoints.kpts)

    def _set_structure(self):

        if self.mode == KpointsMode.manual_set:
            self.corresponding_structure = self.initial_structure

        elif self.mode == KpointsMode.primitive_uniform:
            d = {"structure": self.initial_structure,
                 "symprec": self.symprec,
                 "angle_tolerance": self.angle_tolerance}

            self.corresponding_structure, _ = find_spglib_primitive(**d)
            sym_dataset = get_symmetry_dataset(**d)
            self.sg = sym_dataset["number"]
            self.sg_symbol = sym_dataset["international"]

        elif self.mode == KpointsMode.band:
            lattice = self.seekpath_info["primitive_lattice"]
            element_types = self.seekpath_info["primitive_types"]
            species = [Element.from_Z(i) for i in element_types]
            positions = self.seekpath_info["primitive_positions"]

            self.corresponding_structure = \
                Structure(lattice, species, positions)
            self.sg = self.seekpath_info["spacegroup_number"]
            self.sg_symbol = self.seekpath_info["spacegroup_international"]

        else:
            raise ValueError(f"{self.mode} mode is not supported yet.")

    def _set_kmesh(self):
        self.comment = f"Generated by vise. "
        if self.manual_kpts:
            self.comment += f"Mode: {self.mode}, manual kpt setting. "
            self.kpt_mesh = self.manual_kpts

        else:
            reciprocal_lat = \
                self.corresponding_structure.lattice.reciprocal_lattice
            reciprocal_abc = reciprocal_lat.abc
            if self.mode in (KpointsMode.band, KpointsMode.primitive_uniform):

                body_centered_ortho = {23, 24, 44, 45, 46, 71, 72, 73, 74}
                body_centered_tetra = {79, 80, 81, 82, 87, 88, 97, 98, 107, 108,
                                       109, 110, 119, 120, 121, 122, 139, 140,
                                       141, 142}
                if self.sg in body_centered_ortho | body_centered_tetra:
                    logger.warning("To keep the space group symmetry, the "
                                   "number of k-points along three directions "
                                   "are kept the same for oI and tI Bravais "
                                   "lattice.")
                    average_abc = pow(np.prod(reciprocal_lat.abc), 1 / 3)
                    reciprocal_abc = (average_abc, average_abc, average_abc)

            if self.only_even:
                self.kpt_mesh = [int(ceil(self.kpt_density * r / 2) * 2)
                                 for r in reciprocal_abc]
            else:
                self.kpt_mesh = [int(ceil(self.kpt_density * r))
                                 for r in reciprocal_abc]

            self.kpt_mesh = [i * self.factor for i in self.kpt_mesh]

            self.comment += f"Mode: {self.mode}, " \
                            f"kpt density: {self.kpt_density}, " \
                            f"factor: {self.factor}. "

    def _set_centering(self):
        self.kpt_shift = []
        if self.mode is KpointsMode.manual_set:
            angle = self.corresponding_structure.lattice.angles

            for i in range(3):
                # shift kpt mesh center only for the lattice vector being normal
                # to a lattice plane and even number of k-points.
                if max([abs(angle[i - 2] - 90), abs(angle[i - 1] - 90)]) < 1e-5:
                    self.kpt_shift.append(0.5)
                else:
                    self.kpt_shift.append(0.0)
        else:
            self.kpt_shift = kpt_centering[self.sg]

        # If number of k-point is odd, off centering has no meaning.
        for i in range(3):
            if self.kpt_mesh[i] % 2 == 1:
                self.kpt_shift[i] = 0.0

    def _add_band_kpts(self):
        k_path = self.seekpath_info["explicit_kpoints_rel"]
        k_path_label = self.seekpath_info["explicit_kpoints_labels"]

        formula = self.corresponding_structure.composition.reduced_formula
        self.kpoints.comment += \
            f"k-path added by seekpath. Formula: {formula} SG: {self.sg}"
        self.kpoints.num_kpts += len(k_path)
        self.kpoints.kpts += list(k_path)
        self.kpoints.labels += k_path_label
        self.kpoints.kpts_weights += [0] * len(k_path)

    @property
    def is_structure_changed(self):
        return not np.allclose(self.initial_structure.lattice.matrix,
                               self.corresponding_structure.lattice.matrix,
                               atol=self.symprec)


def irreducible_kpoints(structure: Structure,
                        kpoints: Kpoints,
                        symprec: float = SYMMETRY_TOLERANCE,
                        angle_tolerance: float = ANGLE_TOL) -> List[tuple]:
    """Estimate the irreducible k-points at the given k-mesh written in KPOINTS

    Args:
        kpoints (Kpoints):
            Kpoints file. Supported modes are only Gamma and Monkhorst.
        structure (Structure):
            Input structure corresponding to kpoints.
        symprec (float):
        angle_tolerance (float):
            See docstrings in spglib.

    Returns:
        List of irreducible kpoints and their weights tuples
        [(ir_kpoint, weight)], with ir_kpoint given in fractional coordinates.
    """
    if kpoints.style == Kpoints.supported_modes.Monkhorst:
        kpts_shift = [0.5, 0.5, 0.5]
    elif kpoints.style == Kpoints.supported_modes.Gamma:
        kpts_shift = [0.0, 0.0, 0.0]
    else:
        raise ValueError("Only Gamma or Monkhorst k-points modes supported.")

    kpts_shift = [kpts_shift[x] + kpoints.kpts_shift[x] for x in range(3)]
    # modf(x) returns (fraction part, integer part)
    shift = [1 if modf(i)[0] == 0.5 else 0 for i in kpts_shift]

    sga = SpacegroupAnalyzer(structure=structure,
                             symprec=symprec,
                             angle_tolerance=angle_tolerance)

    return sga.get_ir_reciprocal_mesh(mesh=kpoints.kpts[0], is_shift=shift)


def num_irreducible_kpoints(kpoints: Kpoints,
                            structure: Structure = None,
                            symprec: float = SYMMETRY_TOLERANCE,
                            angle_tolerance: float = ANGLE_TOL) -> int:
    """Calculate number of k-points

    Args:
        kpoints (Kpoints):
            Kpoints file. Supported modes are only Reciprocal, Gamma and
            Monkhorst.
        structure (Structure):
            Input structure corresponding to kpoints.
        symprec (float):
        angle_tolerance (float):
            See docstrings in spglib.

    Return:
        Int of Number of k-points
    """
    if kpoints.style == Kpoints.supported_modes.Reciprocal:
        return len(kpoints.kpts)
    else:
        return len(irreducible_kpoints(structure=structure,
                                       kpoints=kpoints,
                                       symprec=symprec,
                                       angle_tolerance=angle_tolerance))




# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import numpy as np

from pymatgen.analysis.structure_matcher import StructureMatcher

from vise.util.structure_handler import (
    get_symmetry_dataset, structure_to_spglib_cell, spglib_cell_to_structure,
    find_hpkot_primitive, structure_to_seekpath)
from vise.util.testing import ViseTest
from vise.util.error_classes import InvalidStructureError


class GetSymmetryDatasetTest(ViseTest):

    def setUp(self):
        self.structure = self.get_structure_by_name("MgSe300atoms")

    def test(self):
        dataset = get_symmetry_dataset(self.structure, symprec=0.1)
        print(dataset)


class StructureToSpglibCellTest(ViseTest):

    def setUp(self):
        structure = self.get_structure_by_name("conventional_MgO")
        self.cell = structure_to_spglib_cell(structure)

    def test(self):
        expected_lattice = \
            [[4.2468930196252002, 0.0000000000000000, 0.0000000000000000],
             [0.0000000000000000, 4.2468930196252002, 0.0000000000000000],
             [0.0000000000000000, 0.0000000000000000, 4.2468930196252002]]
        expected_1st_position = [0, 0, 0]
        expected_atomic_numbers = [12, 12, 12, 12, 8, 8, 8, 8]

        self.assertTrue(expected_lattice, self.cell[0])
        self.assertTrue(expected_1st_position, self.cell[1][0])
        self.assertTrue(expected_atomic_numbers, self.cell[2])


class SpglibCellToStructureTest(ViseTest):
    def setUp(self):
        self.structure = self.get_structure_by_name("conventional_MgO")
        self.cell = structure_to_spglib_cell(self.structure)
        self.returned_structure = spglib_cell_to_structure(self.cell)

    def test(self):
        self.assertTrue(self.structure == self.returned_structure)


class FindSpglibStandardConventionalTest(ViseTest):
    def setUp(self):
        self.structure = self.get_structure_by_name("MgO")
        self.cell = structure_to_spglib_cell(self.structure)


class FindHPKOTPrimitiveTest(ViseTest):
    def test(self):
        expected = self.get_structure_by_name("MgO")
        actual = find_hpkot_primitive(
            self.get_structure_by_name("conventional_MgO"))

        sm = StructureMatcher(ltol=0.0001, stol=0.0001, angle_tol=0.001,
                              primitive_cell=False, scale=False)
        self.assertTrue(sm.fit(expected, actual))


class StructureToSeekpathTest(ViseTest):

    def setUp(self):
        # self.conventional_cell_structure = \
        #     Structure.from_file("PPOSCAR-MgO-strange_lattice")
        self.primitive_cell_structure = self.get_structure_by_name("MgO")

    # def test_raise_error(self):
    #     with self.assertRaises(InvalidStructureError):
    #         structure_to_seekpath(self.conventional_cell_structure)

    def test_return_res(self):
        res = structure_to_seekpath(self.primitive_cell_structure)
        expected = np.array([0.5, 0, 0.5])
        np.testing.assert_equal(expected, res["explicit_kpoints_rel"][-1])
        self.assertEqual("X", res["explicit_kpoints_labels"][-1])






# -*- coding: utf-8 -*-

from inspect import signature, _empty
from pathlib import Path
import re
from typing import Optional, Callable, List, Union, Tuple

import yaml

from pymatgen.core.periodic_table import Element

from vise.util.tools import is_str_int, is_str_digit, str2bool
from distutils.util import strtobool


def potcar_str2dict(potcar_list: Union[str, List[str], None]) -> dict:
    """Sanitize potcar names to dict.

    If potcar_list is None, {} is returned.
    Examples
        ["Mg_pv", O_h"] -> {"Mg": "Mg_pv", "O": "O_h"}
        "Mg_pv" -> {"Mg": "Mg_pv"}
        None -> {}

    Args:
         potcar_list (str/list/None):
            List of Potcar names or a name, including element names before the
            hyphen.

    Returns:
         Dictionary with element names as keys and potcar names as values.
    """
    if potcar_list is None:
        return {}
    elif isinstance(potcar_list, str):
        potcar_list = [potcar_list]

    d = {}
    for p in potcar_list:
        element = p.split("_")[0]
        if element in d:
            raise ValueError("Multiple POTCAR files for an element are not "
                             "supported yet.")
        try:
            Element(element)
        except ValueError:
            raise ValueError(f"First part of POTCAR file name {p} is not an "
                             f"element supported yet.")

        d[element] = p
    return d


def list2dict(flattened_list: Optional[list], key_candidates: list) -> dict:
    """Sanitize a list to a dict with keys in the flags

    If a string in list does not exist in key_candidates, raise ValueError.

    Examples:
        flattened_list = ["ENCUT", "500", "MAGMOM", "4", "4", "LWAVE", "F"]
        key_candidates = ["ENCUT", "MAGMOM", "LWAVE", ...]
        list2dict(flattened_list, key_candidates) =
                            {"ENCUT": 500, "MAGMOM": [4, 4], "LWAVE": False}

        flattened_list = ["ENCUT", "500", "MAGMAM", "4", "4"]
        list2dict(flattened_list, key_candidates) raises ValueError

    Args:
        flattened_list (list):
            Input list.
        key_candidates (list):
            List of key candidates, e.g., INCAR flags.

    Returns:
        Sanitized dict.
    """
    flattened_list = flattened_list or []

    d = {}
    key = None  # key is None at the beginning or after inserting a value.
    value_list = []

    def insert():
        if not value_list:
            raise ValueError(f"Invalid input: {flattened_list}.")
        d[key] = value_list[0] if len(value_list) == 1 else value_list

    for string in flattened_list:
        if key is None and string not in key_candidates:
            raise ValueError(f"Keys are invalid: {flattened_list}.")
        elif string in key_candidates:
            if key:
                insert()
                key = None
                value_list = []
            key = string
        else:
            # We first need to check numbers as 0 and 1 are parse as False and
            # True with str2bool.
            if is_str_int(string):
                value = int(string)
            else:
                try:
                    value = str2bool(string)
                except ValueError:
                    if is_str_digit(string):
                        value = float(string)
                    else:
                        value = string

            value_list.append(value)
    else:
        if key:
            insert()

    return d


def dict2list(d: dict) -> list:
    """Flatten dict to list w/ string keys. String is also separated by space.

    Example:
        dict2list({"a": 1, "b": "2 3 4", "c": True}) =
                                 ["a", "1", "b", "2", "3", "4", "c", "True"]

    Args:
         d (dict):
            Input dict.

    Returns:
         List of flattened dict.
    """

    d = d or {}
    flattened_list = []
    for k, v in d.items():
        flattened_list.append(k)
        if isinstance(v, str):
            flattened_list.extend(v.split())
        else:
            flattened_list.append(str(v))

    return flattened_list


def get_user_settings(yaml_filename: str,
                      setting_keys: list) -> Tuple[dict, Optional[Path]]:
    """Get the user specifying settings written in yaml_filename

    Note1: The yaml_filename is explored in the parent folders up to home
           or root directory until it's found. If it does not exist, empty
           dictionary is returned.
    Note2: When the key includes "/", the absolute path is added as a prefix.
           E.g., unitcell/unitcell.json -> /something/../unitcell/unitcell.json
    Note3: The value of "potcar_set: Mg_pv O_h" is string of  "Mg_pv O_h", which
           is suited for main default value.

    Args:
        yaml_filename (str):
            User setting yaml filename.
        setting_keys (list):
            Only setting_keys are valid as input keys, otherwise raise
            ValueError.

    Returns:
        Tuple of dictionary of configs and Path to the config file.
    """

    config_path = Path.cwd()
    home = Path.home()

    while True:
        if config_path == home or config_path == Path("/"):
            return {}, None

        f = config_path / yaml_filename
        if f.exists():
            with open(f, "r") as fin:
                user_settings = yaml.load(fin, Loader=yaml.FullLoader)
            break

        else:
            config_path = config_path.parent

    user_settings = user_settings or {}

    # Add full path
    for k, v in user_settings.items():
        if k not in setting_keys:
            raise ValueError(f"Key {k} in {yaml_filename} is invalid."
                             f"The candidate keys are {setting_keys}")
        if isinstance(v, str) and re.match(r'\S*/\S*', v):
            user_settings[k] = str(config_path / v)

    return user_settings, f


def get_default_args(function: Callable) -> dict:
    """Get the default values of the arguments in the method/function.

    inspect._empty means no default.

    Args:
        function (Callable):
            Method or function. when class is inserted, cls.__init__ is called.

    Returns:
        Dict of default values.
    """
    defaults = {}
    signature_obj = signature(function)
    for name, param in signature_obj.parameters.items():
        if param.default != _empty:
            defaults[name] = param.default

    return defaults



# -*- coding: utf-8 -*-

from pymatgen.io.vasp import inputs as orig_inputs
from vise.input_set.incar import ViseIncar

# Monkey patch
orig_inputs.Incar = ViseIncar
ViseVaspInput = orig_inputs.VaspInput



# -*- coding: utf-8 -*-
from copy import deepcopy

from pymatgen import Composition
from pymatgen.io.vasp import Potcar

from vise.input_set.settings_incar import (
    TaskIncarSettings, XcIncarSettings, CommonIncarSettings)
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.util.testing import ViseTest

__author__ = "Yu Kumagai"
__maintainer__ = "Yu Kumagai"


class TaskIncarSettingsTest(ViseTest):
    def setUp(self) -> None:
        mgo = self.get_structure_by_name("MgO")
        self.default_kwargs = {"task": Task.structure_opt,
                               "structure": mgo,
                               "potcar": Potcar(["Mg", "O"]),
                               "num_kpoints": 8,
                               "max_enmax": 400.0,
                               "is_magnetization": False,
                               "vbm_cbm": [3.0, 8.0],
                               "npar_kpar": True,
                               "num_nodes": 1,
                               "encut": None,
                               "structure_opt_encut_factor": 1.3,
                               "dos_step_size": 0.01}

    def test_structure_opt(self):
        setting = TaskIncarSettings.from_options(**self.default_kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 3,
                    'ISMEAR': 0,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-07,
                    'NSW': 50,
                    'EDIFFG': -0.005,
                    'KPAR': 4,
                    'ENCUT': 520.0}
        self.assertEqual(expected, setting.settings)

    def test_structure_opt_rough(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.structure_opt_rough
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 3,
                    'ISMEAR': 0,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-04,
                    'NSW': 50,
                    'EDIFFG': -0.2,
                    'KPAR': 4,
                    'POTIM': 0.1,
                    'ENCUT': 520.0}
        self.assertEqual(expected, setting.settings)

    def test_structure_opt_tight(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.structure_opt_tight
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'ADDGRID': True,
                    'IBRION': 2,
                    'PREC': 'A',
                    'ISIF': 3,
                    'ISMEAR': 0,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-08,
                    'NSW': 50,
                    'EDIFFG': -0.001,
                    'KPAR': 4,
                    'ENCUT': 520.0}
        self.assertEqual(expected, setting.settings)

    def test_cluster_opt(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.cluster_opt
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 2,
                    'ISMEAR': 0,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-07,
                    'NSW': 50,
                    'EDIFFG': -0.005,
                    'KPAR': 4,
                    'ENCUT': 400.0}
        self.assertEqual(expected, setting.settings)

    def test_phonon_force(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.phonon_force
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'ADDGRID': True,
                    'IBRION': 2,
                    'PREC': 'A',
                    'ISIF': 2,
                    'ISMEAR': 0,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-08,
                    'KPAR': 4,
                    'ENCUT': 400.0}
        self.assertEqual(expected, setting.settings)

    def test_defect(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.defect
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 2,
                    'ISMEAR': 0,
                    'ISPIN': 2,
                    'LREAL': "A",
                    'EDIFF': 1e-05,
                    'NSW': 50,
                    'EDIFFG': -0.03,
                    'KPAR': 4,
                    'ENCUT': 400.0}
        self.assertEqual(expected, setting.settings)

    def test_band(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.band
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 0,
                    'ISMEAR': 0,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-05,
                    'KPAR': 4,
                    'NBANDS': 12,
                    'ENCUT': 400.0}
        self.assertEqual(expected, setting.settings)

    def test_dos(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.dos
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 0,
                    'ISMEAR': -5,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-05,
                    'KPAR': 4,
                    'NBANDS': 12,
                    'ENCUT': 400.0,
                    'EMIN': -12.01,
                    'EMAX': 23,
                    'NEDOS': 3502}
        self.assertEqual(expected, setting.settings)

    def test_dielectric_dfpt(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.dielectric_dfpt
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 8,
                    'PREC': 'N',
                    'ISIF': 0,
                    'ISMEAR': 0,
                    'ISPIN': 1,
                    'LREAL': False,
                    'LEPSILON': True,
                    'EDIFF': 1e-06,
                    'KPAR': 4,
                    'ENCUT': 400.0}
        self.assertEqual(expected, setting.settings)

    def test_dielectric_finite_field(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.dielectric_finite_field
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 6,
                    'PREC': 'N',
                    'ISIF': 0,
                    'ISMEAR': 0,
                    'ISPIN': 1,
                    'LCALCEPS': True,
                    'POTIM': 0.015,
                    'LREAL': False,
                    'EDIFF': 1e-06,
                    'KPAR': 4,
                    'ENCUT': 400.0}
        self.assertEqual(expected, setting.settings)

    def test_dielectric_function(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.dielectric_function
        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 0,
                    'ISMEAR': -5,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-05,
                    'KPAR': 4,
                    'NBANDS': 12,
                    'CSHIFT': 0.01,
                    'LOPTICS': True,
                    'ENCUT': 400.0,
                    'EMIN': -12.01,
                    'EMAX': 23,
                    'NEDOS': 3502}
        self.assertEqual(expected, setting.settings)

    def test_args(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.dos
        kwargs["num_kpoints"] = 1
        kwargs["is_magnetization"] = True
        kwargs["vbm_cbm"] = None
        kwargs["encut"] = 800.0

        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 0,
                    'ISMEAR': 0,
                    'ISPIN': 2,
                    'LREAL': False,
                    'EDIFF': 1e-05,
                    'KPAR': 1,
                    'NBANDS': 12,
                    'ENCUT': 800.0,
                    'EMIN': -20.01,
                    'EMAX': 20,
                    'NEDOS': 4002}
        self.assertEqual(expected, setting.settings)

    def test_args_2(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["task"] = Task.dos
        kwargs["num_kpoints"] = 18
        kwargs["num_nodes"] = 2
        kwargs["structure_opt_encut_factor"] = 1.5

        setting = TaskIncarSettings.from_options(**kwargs)
        expected = {'IBRION': 2,
                    'PREC': 'N',
                    'ISIF': 0,
                    'ISMEAR': -5,
                    'ISPIN': 1,
                    'LREAL': False,
                    'EDIFF': 1e-05,
                    'KPAR': 6,
                    'NBANDS': 12,
                    'ENCUT': 400.0,
                    'EMIN': -12.01,
                    'EMAX': 23,
                    'NEDOS': 3502}
        self.assertEqual(expected, setting.settings)


class XcIncarSettingsTest(ViseTest):
    def setUp(self) -> None:
        self.default_kwargs = {"xc": Xc.pbe,
                               "symbol_list": ["Mg", "O"],
                               "factor": 1}

    def test(self):
        setting = XcIncarSettings.from_options(**self.default_kwargs)
        expected = {'LWAVE': False, 'ALGO': 'N'}
        self.assertEqual(expected, setting.settings)

    def test_hubbard_u(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["symbol_list"] = ["Zn", "O"]
        setting = XcIncarSettings.from_options(**kwargs)
        expected = {'ALGO': 'N',
                    'LWAVE': False,
                    'LDAUU': [5, 0],
                    'LDAU': True,
                    'LDAUTYPE': 2,
                    'LDAUPRINT': 1,
                    'LDAUL': [2, -1],
                    'LMAXMIX': 4}
        self.assertEqual(expected, setting.settings)

    def test_hybrid(self):
        kwargs = deepcopy(self.default_kwargs)
        kwargs["xc"] = Xc.hse
        kwargs["factor"] = 2
        setting = XcIncarSettings.from_options(**kwargs)
        expected = {'ALGO': 'D',
                    'LWAVE': True,
                    'NKRED': 2,
                    'HFSCREEN': 0.208,
                    'TIME': 0.4,
                    'LHFCALC': True,
                    'PRECFOCK': 'Fast',
                    'AEXX': 0.25}
        self.assertEqual(expected, setting.settings)


class CommonIncarSettingsTest(ViseTest):
    def setUp(self) -> None:
        self.default_kwargs = {"potcar": Potcar(["Mg", "O"]),
                               "composition": Composition({"Mg": 1, "O": 2}),
                               "charge": 1}

    def test(self):
        setting = CommonIncarSettings.from_options(**self.default_kwargs)
        print(setting.settings)
        expected = {'NELM': 100,
                    'LASPH': True,
                    'LORBIT': 12,
                    'LCHARG': False,
                    'SIGMA': 0.1,
                    'NELECT': 13}
        self.assertEqual(expected, setting.settings)



# -*- coding: utf-8 -*-
#  Copyright (c) 2020. Distributed under the terms of the MIT License.

import os
from argparse import Namespace
from copy import deepcopy
from itertools import chain
from pathlib import Path
from typing import Tuple

from custodian.custodian import Custodian

from pymatgen.analysis.phase_diagram import PhaseDiagram, PDPlotter
from pymatgen.core.periodic_table import Element
from pymatgen.core.structure import Structure
from pymatgen.ext.matproj import MPRester

from vise.analyzer.band_gap import band_gap_properties
from vise.analyzer.band_plotter import PrettyBSPlotter
from vise.analyzer.dos_plotter import get_dos_plot
from vise.chempotdiag.chem_pot_diag import ChemPotDiag
from vise.chempotdiag.free_energy_entries import FreeEnergyEntrySet
from vise.custodian_extension.handler_groups import handler_group
from vise.custodian_extension.jobs import (
    ViseVaspJob, KptConvResult, StructureOptResult)
from vise.input_set.incar import incar_flags
from vise.input_set.input_set import ViseInputSet
from vise.input_set.prior_info import PriorInfo
from vise.input_set.task import Task
from vise.input_set.xc import Xc
from vise.util.error_classes import NoVaspCommandError
from vise.util.logger import get_logger
from vise.cli.main_tools import potcar_str2dict, list2dict
from vise.util.mp_tools import make_poscars_from_mp

logger = get_logger(__name__)


def vasp_settings_from_args(args: Namespace
                            ) -> Tuple[dict, dict]:
    """Generate vasp input settings from the given args.

    """

    flags = [str(s) for s in list(Element)]
    ldauu = list2dict(args.ldauu, flags)
    ldaul = list2dict(args.ldaul, flags)
    potcar_set = potcar_str2dict(args.potcar_set)
    key_candidates = list(ViseInputSet.ALL_OPTIONS.keys())
    # Sanitize values of vis_base_kwargs and user_incar_settings with list2dict.
    vis_base_kwargs = list2dict(args.vise_opts, key_candidates)
    key_candidates = list(chain.from_iterable(incar_flags.values()))
    user_incar_settings = list2dict(args.user_incar_settings, key_candidates)

    if args.additional_user_incar_settings:
        d = list2dict(args.additional_user_incar_settings, key_candidates)
        user_incar_settings.update(d)

    vis_base_kwargs.update({"potcar_set_name": args.potcar_set_name,
                            "override_potcar_set": potcar_set,
                            "ldauu": ldauu,
                            "ldaul": ldaul,
                            "charge": args.charge})
    return user_incar_settings, vis_base_kwargs


def get_poscar_from_mp(args: Namespace) -> None:
    if getattr(args, "number", None):
        s = MPRester().get_structure_by_material_id(f"mp-{args.number}")
        s.to(filename=args.poscar)
    elif getattr(args, "elements", None):
        make_poscars_from_mp(elements=args.elements,
                             e_above_hull=args.e_above_hull,
                             molecules=args.molecules)
    else:
        logger.warning("Set mp number or elements")


def vasp_set(args: Namespace) -> None:
    if args.json:
        vis = ViseInputSet.load_json(args.json)
        print(vis)
        return

    # user_incar_settings and base_vis_kwargs can be modified by prior_info.json
    base_user_incar_settings, base_vis_kwargs = vasp_settings_from_args(args)
    task = Task.from_string(args.task)
    xc = Xc.from_string(args.xc)
    base_vis_kwargs.update(
        {"symprec": args.symprec,
         "angle_tolerance": args.angle_tolerance,
         "kpt_density": args.kpt_density,
         "standardize_structure": args.standardize_structure})

    started_dir = os.getcwd()
    dirs = getattr(args, "dirs", ".")

    for d in dirs:
        d = Path.cwd() / d
        os.chdir(d)
        logger.info(f"Constructing vasp set in {d}")
        user_incar_settings = deepcopy(base_user_incar_settings)
        kwargs = deepcopy(base_vis_kwargs)

        if args.prior_info:
            yaml_file = d / "prior_info.yaml"
            if yaml_file.exists():
                prior_info = PriorInfo.load_yaml(str(yaml_file))
                if prior_info.is_magnetic:
                    kwargs["is_magnetization"] = True
                if prior_info.is_cluster:
                    if task != Task.cluster_opt:
                        logger.warning(f"task is changed from {task} to "
                                       f"{Task.cluster_opt}.")
                        task = Task.cluster_opt
                if prior_info.incar:
                    user_incar_settings.update(prior_info.incar)

        if args.prev_dir:
            logger.warning("user_incar_settings is not used when using "
                           "previous calculation.")
            files = {"CHGCAR": "C", "WAVECAR": "M", "WAVEDER": "M"}
            input_set = ViseInputSet.from_prev_calc(args.prev_dir,
                                                    task=task,
                                                    xc=xc,
                                                    files_to_transfer=files,
                                                    **kwargs)
        else:
            s = Structure.from_file(args.poscar)
            input_set = \
                ViseInputSet.make_input(structure=s,
                                        task=task,
                                        xc=xc,
                                        user_incar_settings=user_incar_settings,
                                        **kwargs)

        input_set.write_input(".")

    os.chdir(started_dir)


def vasp_run_parser(args) -> tuple:
    if isinstance(args.vasp_cmd, str):
        vasp_cmd = args.vasp_cmd.split()
    elif isinstance(args.vasp_cmd, list):
        if len(args.vasp_cmd) == 1:
            vasp_cmd = args.vasp_cmd[0].split()
        else:
            vasp_cmd = args.vasp_cmd
    else:
        raise NoVaspCommandError("Vasp command must be specified properly.")

    handlers = handler_group(args.handler_name, timeout=args.timeout)

    # Used in structure_opt
    optimization_args = {"vasp_cmd": vasp_cmd,
                         "max_relax_num": args.max_relax_num,
                         "removes_wavecar": args.remove_wavecar,
                         "left_files": args.left_files,
                         "removed_files": ["PCDAT", "vasprun.xml"],
                         "symprec": args.symprec,
                         "angle_tolerance": args.angle_tolerance}

    # Used in custodian
    custodian_args = {"handlers": handlers,
                      "polling_time_step": 5,
                      "monitor_freq": 1,
                      "max_errors": 10,
                      "gzipped_output": False}

    return optimization_args, custodian_args


def vasp_run(args) -> None:

    if args.print:
        print(StructureOptResult.load_json(args.json_file))
        return

    optimization_args, custodian_args = vasp_run_parser(args)

    custodian_args["jobs"] = ViseVaspJob.structure_optimization_run(
        **optimization_args)

    c = Custodian(**custodian_args)
    c.run()


def kpt_conv(args) -> None:

    if args.print:
        print(KptConvResult.load_json(args.json_file))
        return

    user_incar_settings, vis_kwargs = vasp_settings_from_args(args)
    optimization_args, custodian_args = vasp_run_parser(args)

    custodian_args["jobs"] = ViseVaspJob.kpt_converge(
        xc=Xc.from_string(args.xc),
        task=Task.from_string(args.task),
        convergence_criterion=args.convergence_criterion,
        initial_kpt_density=args.initial_kpt_density,
        user_incar_settings=user_incar_settings,
        **optimization_args, **vis_kwargs)

    c = Custodian(**custodian_args)
    c.run()


def chempotdiag(args: Namespace) -> None:
    if args.elements:
        entry_set = FreeEnergyEntrySet.from_mp(args.elements)
    else:
        entry_set = FreeEnergyEntrySet.from_vasp_files(
            directory_paths=args.vasp_dirs,
            vasprun=args.vasprun,
            parse_gas=args.parse_gas,
            temperature=args.temperature,
            partial_pressures=args.partial_pressures)

    entry_set.to_json()
    pd = PhaseDiagram(entries=entry_set.entries)

    if args.draw_phase_diagram:
        pd_plotter = PDPlotter(pd)
        if args.filename:
            image_format = args.filename.split(".")[-1]
            pd_plotter.write_image(args.pd_filename, image_format)
        else:
            pd_plotter.show()
        return

    cpd = ChemPotDiag.from_phase_diagram(pd, target_comp=args.target_comp)
    cpd.to_json_file()

    cpd.draw_diagram(filename=args.filename)


def plot_band(args) -> None:
    p = PrettyBSPlotter.from_vasp_files(kpoints_filenames=args.kpoints,
                                        vasprun_filenames=args.vasprun,
                                        vasprun2_filenames=args.vasprun2,
                                        absolute=args.absolute,
                                        y_range=args.y_range,
                                        legend=args.legend,
                                        symprec=args.symprec,
                                        angle_tolerance=args.angle_tolerance)

    p.show(args.filename, format_type="pdf")


def plot_dos(args) -> None:
    if args.cbm_vbm:
        if len(args.cbm_vbm) != 2 or args.cbm_vbm[0] < args.cbm_vbm[1]:
            raise ValueError(f"cbm_vbm values {args.cbm_vbm} are not proper.")

    dos = get_dos_plot(vasprun=args.vasprun,
                       cbm_vbm=args.cbm_vbm,
                       pdos_type=args.pdos_type,
                       specific=args.specific,
                       orbital=args.orbital,
                       xlim=args.x_range,
                       ymaxs=args.ymaxs,
                       zero_at_efermi=not args.absolute,
                       legend=args.legend,
                       crop_first_value=args.crop_first_value,
                       symprec=args.symprec,
                       angle_tolerance=args.angle_tolerance)
    dos.savefig(args.filename, format="pdf")


def band_gap(args) -> None:
    try:
        band_gap_value, vbm_info, cbm_info = \
            band_gap_properties(vasprun=args.vasprun, outcar=args.outcar)
        print(f"CBM info {cbm_info}")
        print(f"VBM info {vbm_info}")
        print(f"band gap info {band_gap_value}")
    except TypeError:
        print("Metallic system")





# -*- coding: utf-8 -*-

from pymatgen.io.vasp import Potcar

from vise.input_set.settings_util import load_potcar_yaml
from vise.input_set.xc import Xc


class XcTaskPotcar:

    def __init__(self,
                 potcar: Potcar,
                 max_enmax: int):

        self.potcar = potcar
        self.max_enmax = max_enmax

    @classmethod
    def from_options(cls,
                     xc: Xc,
                     symbol_list: list,
                     potcar_set_name: str = None,
                     override_potcar_set: dict = None):
        """Construct Potcar from xc and some options.

        Args: See ViseInputSet docstrings

        Returns:
            XcTaskPotcar instance object
        """

        potcar_functional = "LDA" if xc == Xc.lda else "PBE_54"
        # Currently GW calc is not supported.
        potcar_set_name = potcar_set_name or "normal"

        potcar_list = load_potcar_yaml(potcar_set_name, override_potcar_set)
        potcar_symbols = [potcar_list.get(el, el) for el in symbol_list]
        potcar = Potcar(potcar_symbols, functional=potcar_functional)

        max_enmax = max([p.enmax for p in potcar])

        return cls(potcar, max_enmax)
